import numpy as np
import pandas as pd
import pytest
from scipy.stats import vonmises

import gwpopulation
from gwpopulation import utils

N_TEST = 100


def test_beta_dist_zero_below_zero():
    equal_zero = True
    for _ in range(N_TEST):
        vals = utils.beta_dist(-1, *np.random.uniform(0, 10, 3))
        equal_zero = equal_zero & (vals == 0.0)
    assert equal_zero


def test_beta_dist_zero_above_scale():
    equal_zero = True
    for _ in range(N_TEST):
        vals = utils.beta_dist(20, *np.random.uniform(0, 10, 3))
        equal_zero = equal_zero & (vals == 0.0)
    assert equal_zero


def test_beta_dist_alpha_below_zero_raises_value_error():
    gwpopulation.set_backend("numpy")
    with pytest.raises(ValueError):
        utils.beta_dist(xx=0.5, alpha=-1, beta=1, scale=1)


def test_beta_dist_beta_below_zero_raises_value_error():
    gwpopulation.set_backend("numpy")
    with pytest.raises(ValueError):
        utils.beta_dist(xx=0.5, alpha=1, beta=-1, scale=1)


def test_powerlaw_zero_below_low():
    equal_zero = True
    for ii in range(N_TEST):
        vals = utils.powerlaw(
            xx=1,
            alpha=np.random.uniform(-10, 10),
            low=np.random.uniform(5, 15),
            high=np.random.uniform(20, 30),
        )
        equal_zero = equal_zero & (vals == 0.0)
    assert equal_zero


def test_powerlaw_zero_above_high():
    equal_zero = True
    for ii in range(N_TEST):
        vals = utils.powerlaw(
            xx=40,
            alpha=np.random.uniform(-10, 10),
            low=np.random.uniform(5, 15),
            high=np.random.uniform(20, 30),
        )
        equal_zero = equal_zero & (vals == 0.0)
    assert equal_zero


def test_powerlaw_low_below_zero_raises_value_error():
    with pytest.raises(ValueError):
        utils.powerlaw(xx=0, alpha=3, high=10, low=-4)


def test_powerlaw_alpha_equal_zero(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    assert utils.powerlaw(xx=1.0, alpha=-1, low=0.5, high=2) == 1 / xp.log(4)
    gwpopulation.set_backend("numpy")


def test_truncnorm_zero_below_low():
    equal_zero = True
    for _ in range(N_TEST):
        vals = utils.truncnorm(
            -40,
            mu=np.random.uniform(-10, 10),
            sigma=np.random.uniform(0, 10),
            low=np.random.uniform(-30, -20),
            high=np.random.uniform(20, 30),
        )
        equal_zero = equal_zero & (vals == 0.0)
    assert equal_zero


def test_truncnorm_zero_above_high():
    equal_zero = True
    for _ in range(N_TEST):
        vals = utils.truncnorm(
            40,
            mu=np.random.uniform(-10, 10),
            sigma=np.random.uniform(0, 10),
            low=np.random.uniform(-30, -20),
            high=np.random.uniform(20, 30),
        )
        equal_zero = equal_zero & (vals == 0.0)
    assert equal_zero


def test_truncnorm_sigma_below_zero_raises_value_error():
    with pytest.raises(ValueError):
        utils.truncnorm(xx=0, mu=0, sigma=-1, high=10, low=-10)


def test_truncnorm_matches_scipy(backend):
    from scipy.stats import truncnorm

    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    xx = xp.linspace(-2, 2, 1000)
    for ii in range(N_TEST):
        mu = np.random.uniform(-10, 10)
        sigma = np.random.uniform(0, 5)
        low, high = np.sort(np.random.uniform(-10, 10, 2))
        gwpop_vals = utils.to_numpy(
            utils.truncnorm(xx, mu=mu, sigma=sigma, low=low, high=high)
        )
        scipy_vals = truncnorm(
            loc=mu, scale=sigma, a=(low - mu) / sigma, b=(high - mu) / sigma
        ).pdf(utils.to_numpy(xx))
        assert max(abs(gwpop_vals - scipy_vals)) < 1e-3


def test_matches_scipy(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    xx = xp.linspace(0, 2 * np.pi, 1000)
    for ii in range(N_TEST):
        mu = np.random.uniform(-np.pi, np.pi)
        kappa = np.random.uniform(0, 15)
        gwpop_vals = utils.to_numpy(utils.von_mises(xx, mu, kappa))
        scipy_vals = vonmises(kappa=kappa, loc=mu).pdf(utils.to_numpy(xx))
        assert max(abs(gwpop_vals - scipy_vals)) < 1e-3


def test_to_numpy_leaves_pandas_changed():
    test = pd.Series([1, 2, 3])
    assert type(test) == type(utils.to_numpy(test))


def test_to_numpy_non_array_type_raises_error():
    with pytest.raises(TypeError):
        utils.to_numpy([1, 2, 3])


@gwpopulation.utils.apply_conditions(dict(a=lambda a: a > 0))
def _condition_func(a):
    return a


def test_callable_condition_not_satisfied():
    gwpopulation.set_backend("numpy")
    with pytest.raises(ValueError):
        _condition_func(a=-1)


def test_callable_condition_satisfied():
    gwpopulation.set_backend("numpy")
    assert _condition_func(a=1) == 1


def test_non_callable_op_raises_error():
    gwpopulation.set_backend("numpy")

    @gwpopulation.utils.apply_conditions(dict(a=("ge", 0)))
    def _condition_func(a):
        return

    with pytest.raises(ValueError):
        _condition_func(a=1)


def test_trapezoid_matches_scipy(backend):
    from scipy.integrate import trapezoid

    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    x = xp.linspace(0, 10, 1000)
    y = xp.sin(x) + 1.1
    gwpop_int = utils.trapezoid(y, x)
    scipy_int = float(trapezoid(y, x))
    assert abs(gwpop_int - scipy_int) < 1e-10



r"""
Gravitational-wave transient surveys provide a biased sample of the astrophysical population.
The likelihood function used for population inference is given be

.. math::

    {\cal L}(\{d_i\} | \Lambda) &= \prod_i {\cal L}(d_i | \Lambda, {\rm det})

    &= \prod_i \frac{{\cal L}(d_i | \Lambda)}{P_{\rm det}(\Lambda)}

    &= \frac{1}{P_{\rm det}^{N}(\Lambda)} \prod_i \int d\theta_i p(d_i | \theta_i) \pi(\theta_i | \Lambda).

The quantity :math:`P_{\rm det}(\Lambda)` is the detection probability for a single source (see `<selection.html>`_).

The integrals over the per-event parameters :math:`\theta_i` are typically performed using Monte Carlo integration

.. math::

    \hat{{\cal L}}(d_i | \Lambda) = \frac{1}{K} \sum_{k=1}^K \frac{\pi(\theta_k | \Lambda)}{\pi(\theta_k | \varnothing)}.

The full approximate log-likelihood is then given by

.. math::

    \ln \hat{\cal L}(\{d_i\} | \Lambda) = \sum_i \ln \hat{{\cal L}}(d_i | \Lambda) - N \ln \hat{P}_{\rm det}(\Lambda).

This approximation is implemented in :class:`gwpopulation.hyperpe.HyperparameterLikelihood`.

There is another related expression for the likelihood as the result of an inhomoegeneous Poisson process.
In this case the likelihood is given by

.. math::

    \ln {\cal L}(\{d_i\} | \Lambda) = N \ln R - N_{\rm exp}(\Lambda)
    + \sum_i \ln \hat{{\cal L}}(d_i | \Lambda)

Here :math:`R` is the total merger rate and :math:`T` is the total observation time
and :math:`N_{\rm exp}(\Lambda) = RT\hat{P}_{\rm det}(\Lambda)`.
This is implemented in :class:`gwpopulation.hyperpe.RateLikelihood`.

Each of these Monte Carlo integrals have associated uncertainties which are propagated through the likelihood calculation
and can be calculated using :func:`gwpopulation.hyperpe.HyperparameterLikelihood.ln_likelihood_and_variance`.
"""

import types

import numpy as np
from bilby.core.likelihood import Likelihood
from bilby.core.utils import logger
from bilby.hyper.model import Model

from .utils import get_name, to_number, to_numpy

xp = np

__all__ = ["HyperparameterLikelihood", "RateLikelihood", "xp"]


class HyperparameterLikelihood(Likelihood):
    """
    A likelihood for inferring hyperparameter posterior distributions with
    including selection effects.

    See Eq. (34) of `Thrane and Talbot <https://arxiv.org/abs/1809.02293>`_
    for a definition.

    For the uncertainty calculation see the Appendix of
    `Golomb and Talbot <https://arxiv.org/abs/2106.15745>`_ and
    `Farr <https://arxiv.org/abs/1904.10879>`_.
    """

    def __init__(
        self,
        posteriors,
        hyper_prior,
        ln_evidences=None,
        max_samples=1e100,
        selection_function=lambda args: 1,
        conversion_function=lambda args: (args, None),
        maximum_uncertainty=xp.inf,
    ):
        """
        Parameters
        ----------
        posteriors: list
            An list of pandas data frames of samples sets of samples.
            Each set may have a different size.
            These can contain a `prior` column containing the original prior
            values.
        hyper_prior: `bilby.hyper.model.Model`
            The population model, this can alternatively be a function.
        ln_evidences: list, optional
            Log evidences for single runs to ensure proper normalisation
            of the hyperparameter likelihood. If not provided, the original
            evidences will be set to 0. This produces a Bayes factor between
            the sampling power_prior and the hyperparameterised model.
        selection_function: func
            Function which evaluates your population selection function.
        conversion_function: func
            Function which converts a dictionary of sampled parameter to a
            dictionary of parameters of the population model.
        max_samples: int, optional
            Maximum number of samples to use from each set.
        maximum_uncertainty: float
            The maximum allowed uncertainty in the natural log likelihood.
            If the uncertainty is larger than this value a log likelihood of
            -inf will be returned. Default = inf
        """

        self.samples_per_posterior = max_samples
        self.data = self.resample_posteriors(posteriors, max_samples=max_samples)

        if isinstance(hyper_prior, types.FunctionType):
            hyper_prior = Model([hyper_prior])
        elif not (
            hasattr(hyper_prior, "parameters")
            and callable(getattr(hyper_prior, "prob"))
        ):
            raise AttributeError(
                "hyper_prior must either be a function, "
                "or a class with attribute 'parameters' and method 'prob'"
            )
        self.hyper_prior = hyper_prior
        super().__init__()

        if "prior" in self.data:
            self.sampling_prior = self.data.pop("prior")
        else:
            logger.info("No prior values provided, defaulting to 1.")
            self.sampling_prior = 1

        if ln_evidences is not None:
            self.total_noise_evidence = np.sum(ln_evidences)
        else:
            self.total_noise_evidence = np.nan

        self.conversion_function = conversion_function
        self.selection_function = selection_function

        self.n_posteriors = len(posteriors)
        self.maximum_uncertainty = maximum_uncertainty
        self._inf = np.nan_to_num(np.inf)

    __doc__ += __init__.__doc__

    @property
    def maximum_uncertainty(self):
        """
        The maximum allowed uncertainty in the estimate of the log-likelihood.
        If the uncertainty is larger than this value a log likelihood of -inf
        is returned.
        """
        return self._maximum_uncertainty

    @maximum_uncertainty.setter
    def maximum_uncertainty(self, value):
        self._maximum_uncertainty = value
        if value in [xp.inf, np.inf]:
            self._max_variance = value
        else:
            self._max_variance = value**2

    def ln_likelihood_and_variance(self, parameters):
        """
        Compute the ln likelihood estimator and its variance.
        """
        parameters, added_keys = self.conversion_function(parameters)
        ln_bayes_factors, variances = self._compute_per_event_ln_bayes_factors(
            parameters
        )
        ln_l = xp.sum(ln_bayes_factors)
        variance = xp.sum(variances)
        selection, selection_variance = self._get_selection_factor(
            parameters=parameters
        )
        variance += selection_variance
        ln_l += selection
        return ln_l, to_number(variance, float)

    def log_likelihood_ratio(self, parameters):
        ln_l, variance = self.ln_likelihood_and_variance(parameters=parameters)
        ln_l = xp.nan_to_num(ln_l, nan=-xp.inf)
        ln_l -= xp.nan_to_num(xp.inf * (self.maximum_uncertainty < variance), nan=0)
        return to_number(xp.nan_to_num(ln_l), float)

    def noise_log_likelihood(self):
        return self.total_noise_evidence

    def log_likelihood(self, parameters):
        return self.noise_log_likelihood() + self.log_likelihood_ratio(
            parameters=parameters
        )

    def _compute_per_event_ln_bayes_factors(
        self, parameters, *, return_uncertainty=True
    ):
        weights = self.hyper_prior.prob(self.data, **parameters) / self.sampling_prior
        expectation = xp.mean(weights, axis=-1)
        if return_uncertainty:
            square_expectation = xp.mean(weights**2, axis=-1)
            variance = (square_expectation - expectation**2) / (
                self.samples_per_posterior * expectation**2
            )
            return xp.log(expectation), variance
        else:
            return xp.log(expectation)

    def _get_selection_factor(self, parameters, *, return_uncertainty=True):
        selection, variance = self._selection_function_with_uncertainty(
            parameters=parameters
        )
        total_selection = -self.n_posteriors * xp.log(selection)
        if return_uncertainty:
            total_variance = self.n_posteriors**2 * xp.divide(variance, selection**2)
            return total_selection, total_variance
        else:
            return total_selection

    def _selection_function_with_uncertainty(self, parameters):
        result = self.selection_function(parameters)
        if isinstance(result, tuple):
            selection, variance = result
        else:
            selection = result
            variance = 0.0
        return selection, variance

    def generate_extra_statistics(self, sample):
        r"""
        Given an input sample, add extra statistics

        Adds:

        - :code:`ln_bf_idx`: :math:`\frac{\ln {\cal L}(d_{i} | \Lambda)}
          {\ln {\cal L}(d_{i} | \varnothing)}`
          for each of the events in the data
        - :code:`selection`: :math:`P_{\rm det}`
        - :code:`var_idx`, :code:`selection_variance`: the uncertainty in
          each Monte Carlo integral
        - :code:`total_variance`: the total variance in the likelihood

        .. note::

            The quantity :code:`selection_variance` is the variance in
            :code:`P_{\rm det}` and not the total variance from the contribution
            of the selection function to the likelihood.

        Parameters
        ----------
        sample: dict
            Input sample to compute the extra things for.

        Returns
        -------
        sample: dict
            The input dict, modified in place.
        """
        parameters, added_keys = self.conversion_function(sample.copy())
        ln_ls, variances = self._compute_per_event_ln_bayes_factors(
            parameters, return_uncertainty=True
        )
        total_variance = sum(variances)
        for ii in range(self.n_posteriors):
            sample[f"ln_bf_{ii}"] = to_number(ln_ls[ii], float)
            sample[f"var_{ii}"] = to_number(variances[ii], float)
        selection, variance = self._selection_function_with_uncertainty(
            parameters=parameters
        )
        variance /= selection**2
        selection_variance = variance * self.n_posteriors**2
        sample["selection"] = selection
        sample["selection_variance"] = variance
        total_variance += selection_variance
        sample["variance"] = to_number(total_variance, float)
        return sample

    def generate_rate_posterior_sample(self, parameters):
        r"""
        Generate a sample from the posterior distribution for rate assuming a
        :math:`1 / R` prior.

        The likelihood evaluated is analytically marginalized over rate.
        However the rate dependent likelihood can be trivially written.

        .. math::
            p(R) = \Gamma(n=N, \text{scale}=\mathcal{V})

        Here :math:`\Gamma` is the Gamma distribution, :math:`N` is the number
        of events being analyzed and :math:`\mathcal{V}` is the total observed 4-volume.

        .. note::

            This function only uses the :code:`numpy` backend. It can be used
            with the other backends as it returns a float, but does not support
            e.g., autodifferentiation with :code:`jax`.

        Returns
        -------
        rate: float
            A sample from the posterior distribution for rate.
        """
        from scipy.stats import gamma

        if hasattr(self.selection_function, "detection_efficiency") and hasattr(
            self.selection_function, "surveyed_hypervolume"
        ):
            efficiency, _ = self.selection_function.detection_efficiency(parameters)
            vt = efficiency * self.selection_function.surveyed_hypervolume(parameters)
        else:
            vt = self.selection_function(parameters)
        rate = gamma(a=self.n_posteriors).rvs() / vt
        return rate

    def resample_posteriors(self, posteriors, max_samples=1e300):
        """
        Convert list of pandas DataFrame object to dict of arrays.

        Parameters
        ----------
        posteriors: list
            List of pandas DataFrame objects.
        max_samples: int, opt
            Maximum number of samples to take from each posterior,
            default is length of shortest posterior chain.

        Returns
        -------
        data: dict
            Dictionary containing arrays of size (n_posteriors, max_samples)
            There is a key for each shared key in posteriors.
        """
        for posterior in posteriors:
            max_samples = min(len(posterior), max_samples)
        data = {key: [] for key in posteriors[0]}
        logger.debug(f"Downsampling to {max_samples} samples per posterior.")
        self.samples_per_posterior = max_samples
        for posterior in posteriors:
            temp = posterior.sample(self.samples_per_posterior)
            for key in data:
                data[key].append(temp[key])
        for key in data:
            data[key] = xp.array(data[key])
        return data

    def posterior_predictive_resample(self, samples, return_weights=False):
        """
        Resample the original single event posteriors to use the PPD from each
        of the other events as the prior.

        Parameters
        ----------
        samples: pd.DataFrame, dict, list
            The samples to do the weighting over, typically the posterior from
            some run.
        return_weights: bool, optional
            Whether to return the per-sample weights, default = :code:`False`

        Returns
        -------
        new_samples: dict
            Dictionary containing the weighted posterior samples for each of
            the events.
        weights: array-like
            Weights to apply to the samples, only if :code:`return_weights == True`.
        """
        import pandas as pd
        from tqdm.auto import tqdm

        if isinstance(samples, pd.DataFrame):
            samples = [dict(samples.iloc[ii]) for ii in range(len(samples))]
        elif isinstance(samples, dict):
            samples = [samples]
        weights = xp.zeros((self.n_posteriors, self.samples_per_posterior))
        event_weights = xp.zeros(self.n_posteriors)
        for sample in tqdm(samples):
            parameters, added_keys = self.conversion_function(sample.copy())
            new_weights = (
                self.hyper_prior.prob(self.data, **parameters) / self.sampling_prior
            )
            event_weights += xp.mean(new_weights, axis=-1)
            new_weights = (new_weights.T / xp.sum(new_weights, axis=-1)).T
            weights += new_weights
        weights = (weights.T / xp.sum(weights, axis=-1)).T
        new_idxs = xp.empty_like(weights, dtype=int)
        for ii in range(self.n_posteriors):
            if "jax" in xp.__name__:
                from jax import random

                rng_key = random.PRNGKey(np.random.randint(10000000))
                new_idxs = new_idxs.at[ii].set(
                    random.choice(
                        rng_key,
                        xp.arange(self.samples_per_posterior),
                        shape=(self.samples_per_posterior,),
                        replace=True,
                        p=weights[ii],
                    )
                )
            else:
                new_idxs[ii] = xp.asarray(
                    np.random.choice(
                        range(self.samples_per_posterior),
                        size=self.samples_per_posterior,
                        replace=True,
                        p=to_numpy(weights[ii]),
                    )
                )
        new_samples = {
            key: xp.vstack(
                [self.data[key][ii, new_idxs[ii]] for ii in range(self.n_posteriors)]
            )
            for key in self.data
        }
        event_weights = list(event_weights)
        weight_string = " ".join([f"{float(weight):.1f}" for weight in event_weights])
        logger.info(f"Resampling done, sum of weights for events are {weight_string}")
        if return_weights:
            return new_samples, weights
        else:
            return new_samples

    @property
    def meta_data(self):
        return dict(
            model=[get_name(model) for model in self.hyper_prior.models],
            data={key: to_numpy(self.data[key]) for key in self.data},
            n_events=self.n_posteriors,
            sampling_prior=to_numpy(self.sampling_prior),
            samples_per_posterior=self.samples_per_posterior,
        )


class RateLikelihood(HyperparameterLikelihood):
    """
    A likelihood for inferring hyperparameter posterior distributions
    and estimating rates with including selection effects.

    See Eq. (34) of `Thrane and Talbot <https://arxiv.org/abs/1809.02293>`_
    for a definition.

    """

    __doc__ += HyperparameterLikelihood.__init__.__doc__

    def _get_selection_factor(self, parameters, *, return_uncertainty=True):
        r"""
        The selection factor for the rate likelihood is

        .. math::

            \ln P_{\rm det} = N \ln R - N_{\rm exp}(\Lambda)

        The uncertainty is given by

        .. math::

            \sigma^2 = \frac{N_{\rm exp}(\Lambda) \sigma^2_{\rm det}}{P_{\rm det}^2}

        Parameters
        ----------
        return_uncertainty: bool
            Whether to return the uncertainty in the selection factor.
        """
        selection, variance = self._selection_function_with_uncertainty(
            parameters=parameters
        )
        n_expected = selection * parameters["rate"]
        total_selection = -n_expected + self.n_posteriors * xp.log(parameters["rate"])
        if return_uncertainty:
            total_variance = n_expected * variance / selection**2
            return total_selection, total_variance
        else:
            return total_selection

    def generate_rate_posterior_sample(self, parameters):
        """
        Since the rate is a sampled parameter,
        this simply returns the current value of the rate parameter.
        """
        return parameters["rate"]


class NullHyperparameterLikelihood(HyperparameterLikelihood):
    """
    A likelihood that can be used to sample the prior space subject to the
    maximum_uncertainty constraint imposed on the variance.

    For the uncertainty calculation see the Appendix of
    `Golomb and Talbot <https://arxiv.org/abs/2106.15745>`_ and
    `Farr <https://arxiv.org/abs/1904.10879>`_.
    """

    def ln_likelihood_and_variance(self, parameters=None):
        """
        Compute the ln likelihood estimator and its variance.
        """
        _, variance = super().ln_likelihood_and_variance(parameters)
        return 0.0, variance



# -- Project information --------------------------------------------------

import importlib.metadata
import os

project = "GWPopulation"
copyright = "2021-2024, Colm Talbot"
author = "Colm Talbot"

# -- General configuration ------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "numpydoc",
    "nbsphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "autoapi.extension",
    "myst_parser",
    "sphinx_design",
    "sphinx_copybutton",
    "sphinx_tabs.tabs",
]

templates_path = ["_templates"]
html_static_path = ["_static"]
html_favicon = "_static/logo.svg"
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "requirements.txt"]

root_doc = "index"
source_suffix = [".rst", ".md"]


# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
fullversion = importlib.metadata.version("gwpopulation")
version = ".".join(fullversion.split(".")[:2])
# The full version, including alpha/beta/rc tags.
release = fullversion.split("+")[0]

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"


# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
# todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

html_theme = "pydata_sphinx_theme"

autosummary_generate = True
numpydoc_show_class_members = False

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}


autoapi_dirs = ["../gwpopulation"]
autoapi_ignore = ["*_version*"]

json_url = "https://wcosmo.readthedocs.io/en/latest/_static/switcher.json"

version_match = os.environ.get("READTHEDOCS_VERSION")
# If READTHEDOCS_VERSION doesn't exist, we're not on RTD
# If it is an integer, we're in a PR build and the version isn't correct.
# If it's "latest" → change to "dev" (that's what we want the switcher to call it)
if not version_match or version_match.isdigit() or version_match == "latest":
    # For local development, infer the version to match from the package.
    if "dev" in release or "rc" in release:
        version_match = "dev"
        # We want to keep the relative reference if we are in dev mode
        # but we want the whole url if we are effectively in a released version
        json_url = "_static/switcher.json"
    else:
        version_match = f"v{release}"
elif version_match == "stable":
    version_match = f"v{release}"

html_theme_options = {
    "logo": {
        "text": f"GWPopulation v{version}",
        "image_light": "_static/logo.svg",
        "image_dark": "_static/logo-dark.svg",
    },
    "navbar_center": ["navbar-nav"],
    # "navbar_start": ["navbar-logo", "version-switcher"],
    "navbar_start": ["navbar-logo"],
    "navbar_end": ["theme-switcher", "navbar-icon-links"],
    "use_edit_page_button": True,
    "show_version_warning_banner": True,
    "footer_start": ["copyright"],
    "footer_center": ["sphinx-version"],
    # "switcher": {
    #     "json_url": json_url,
    #     "version_match": version_match,
    # },
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/ColmTalbot/GWPopulation",
            "icon": "fa-brands fa-github",
        },
    ],
}

html_context = {
    "github_user": "ColmTalbot",
    "github_repo": "gwpopulation",
    "github_version": "main",
    "doc_path": "docs",
}

nbsphinx_execute = "never"
nbsphinx_prolog = """
This page was automatically generated from a Jupyter notebook. Find the original here_.

.. image:: https://colab.research.google.com/assets/colab-badge.svg
    :target: https://colab.research.google.com/github/ColmTalbot/GWPopulation/blob/main/examples/{{ env.doc2path(env.docname, base=None) }}

.. _here: https://github.com/ColmTalbot/gwpopulation/tree/main/examples/{{ env.doc2path(env.docname, base=None) }}

----
"""



"""
Implemented spin models
"""

import numpy as xp

from ..utils import beta_dist, trapezoid, truncnorm, unnormalized_2d_gaussian
from .interped import InterpolatedNoBaseModelIdentical

__all__ = [
    "GaussianChiEffChiP",
    "SplineSpinMagnitudeIdentical",
    "SplineSpinTiltIdentical",
    "iid_spin",
    "iid_spin_magnitude_beta",
    "independent_spin_magnitude_beta",
    "iid_spin_orientation_gaussian_isotropic",
    "independent_spin_orientation_gaussian_isotropic",
    "gaussian_chi_eff",
    "gaussian_chi_p",
]


def iid_spin(dataset, xi_spin, sigma_spin, amax, alpha_chi, beta_chi):
    r"""
    Independently and identically distributed spins.
    The magnitudes are assumed to follow a Beta distribution and the
    orientations are assumed to follow an isotropic + truncated half
    Gaussian mixture model.

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays containing 'a_1' and 'a_2'.
    xi_spin: float
        Fraction of black holes in preferentially aligned component.
    sigma_spin: float
        Width of preferentially aligned component.
    alpha_chi, beta_chi: float
        Parameters of Beta distribution for both black holes.
    amax: float
        Maximum black hole spin.
    """
    prior = iid_spin_orientation_gaussian_isotropic(
        dataset, xi_spin, sigma_spin
    ) * iid_spin_magnitude_beta(dataset, amax, alpha_chi, beta_chi)
    return prior


def iid_spin_magnitude_beta(dataset, amax=1, alpha_chi=1, beta_chi=1):
    """
    Independent and identically distributed beta distributions for both spin magnitudes.

    See `Wysocki+ <https://arxiv.org/abs/1805.06442>` Eq. (10)

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays containing 'a_1' and 'a_2'.
    alpha_chi, beta_chi: float
        Parameters of Beta distribution for both black holes.
    amax: float
        Maximum black hole spin.
    """
    return independent_spin_magnitude_beta(
        dataset, alpha_chi, alpha_chi, beta_chi, beta_chi, amax, amax
    )


def independent_spin_magnitude_beta(
    dataset, alpha_chi_1, alpha_chi_2, beta_chi_1, beta_chi_2, amax_1, amax_2
):
    """
    Independent beta distributions for both spin magnitudes.

    See `Wysocki+ <https://arxiv.org/abs/1805.06442>` Eq. (10)

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays containing 'a_1' and 'a_2'.
    alpha_chi_1, beta_chi_1: float
        Parameters of Beta distribution for more massive black hole.
    alpha_chi_2, beta_chi_2: float
        Parameters of Beta distribution for less massive black hole.
    amax_1, amax_2: float
        Maximum spin of the more/less massive black hole.
    """
    prior = beta_dist(
        dataset["a_1"], alpha_chi_1, beta_chi_1, scale=amax_1
    ) * beta_dist(dataset["a_2"], alpha_chi_2, beta_chi_2, scale=amax_2)
    return prior


def iid_spin_orientation_gaussian_isotropic(dataset, xi_spin, sigma_spin, mu_spin=1):
    r"""
    A mixture model of spin orientations with isotropic and normally
    distributed components. The distribution of primary and secondary spin
    orientations are expected to be identical and independent.

    See `Talbot and Thrane <https://arxiv.org/abs/1704.08370>`_ Eq. (4)

    .. math::

        p(z_1, z_2 | \xi, \sigma) =
        \frac{(1 - \xi)^2}{4}
        + \xi \prod_{i\in\{1, 2\}} \mathcal{N}_{[-1, 1]}(z_i; \mu=1, \sigma=\sigma)

    Where :math:`\mathcal{N}_{[a, b]}` is the truncated normal distribution over :math:`[a, b]`.

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'cos_tilt_1' and 'cos_tilt_2'.
    xi_spin: float
        Fraction of black holes in gaussian component (:math:`\xi`).
    sigma_spin: float
        Width of gaussian component.
    mu_spin: float
        Mean of the gaussian component (default: cos_tilt = 1).
    """
    return independent_spin_orientation_gaussian_isotropic(
        dataset, xi_spin, sigma_spin, sigma_spin, mu_spin, mu_spin
    )


def independent_spin_orientation_gaussian_isotropic(
    dataset, xi_spin, sigma_1, sigma_2, mu_1=1, mu_2=1
):
    r"""
    A mixture model of spin orientations with isotropic and normally
    distributed components.

    See `Talbot and Thrane <https://arxiv.org/abs/1704.08370>`_ Eq. (4)

    .. math::

        p(z_1, z_2 | \xi, \sigma) =
        \frac{(1 - \xi)^2}{4}
        + \xi \prod_{i\in\{1, 2\}} \mathcal{N}_{[-1, 1]}(z_i; \mu=1, \sigma=\sigma_{i})

    Where :math:`\mathcal{N}_{[a, b]}` is the truncated normal distribution over :math:`[a, b]`.

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for 'cos_tilt_1' and 'cos_tilt_2'.
    xi_spin: float
        Fraction of black holes in gaussian component (:math:`\xi`).
    mu_1: float
        Mean of the gaussian component for the more
        massive black hole (:math:`\mu_1`).
    mu_2: float
        Mean of the gaussian component for the less
        massive black hole (:math:`\mu_2`).
    sigma_1: float
        Width of the gaussian component for the more
        massive black hole (:math:`\sigma_1`).
    sigma_2: float
        Width of the gaussian component for the less
        massive black hole (:math:`\sigma_2`).
    """
    prior = (1 - xi_spin) / 4 + xi_spin * truncnorm(
        dataset["cos_tilt_1"], mu_1, sigma_1, 1, -1
    ) * truncnorm(dataset["cos_tilt_2"], mu_2, sigma_2, 1, -1)
    return prior


def gaussian_chi_eff(dataset, mu_chi_eff, sigma_chi_eff):
    r"""
    A Gaussian in chi effective distribution


    .. math::
        p(\chi_{\text{eff}} | \mu_\chi, \sigma_\chi) =
        \mathcal{N}_{[-1, 1]}(\chi_{\text{eff}}; \mu=\mu_\chi, \sigma=\sigma_\chi)

    Where :math:`\mathcal{N}_{[a, b]}` is the truncated normal distribution over :math:`[a, b]`.

    See `Miller+ <https://arxiv.org/abs/2001.06051>`_ and `Callister+ <https://arxiv.org/abs/2010.14533>`_.

    Parameters
    ----------
    dataset: dict
        Input data, must contain `chi_eff` (:math:`\chi_{\text{eff}}`)
    mu_chi_eff: float
        Mean of the distribution (:math:`\mu_\chi`)
    sigma_chi_eff: float
        Standard deviation of the distribution (:math:`\sigma_\chi`)

    Returns
    -------
    array-like: The probability
    """
    return truncnorm(
        dataset["chi_eff"], mu=mu_chi_eff, sigma=sigma_chi_eff, low=-1, high=1
    )


def gaussian_chi_p(dataset, mu_chi_p, sigma_chi_p):
    r"""
    A Gaussian distribution in precessing effective spin (chi p)

    .. math::
        p(\chi_p) = \mathcal{N}_{[0, 1]}(\chi_p; \mu=\mu_\chi, \sigma=\sigma_\chi)

    Where :math:`\mathcal{N}_{[a, b]}` is the truncated normal distribution over :math:`[a, b]`.

    See `Miller+ <https://arxiv.org/abs/2001.06051>`_ and `Callister+ <https://arxiv.org/abs/2010.14533>`_.

    Parameters
    ----------
    dataset: dict
        Input data, must contain `chi_eff` (:math:`\chi_p`)
    mu_chi_p: float
        Mean of the distribution (:math:`\mu_\chi`)
    sigma_chi_p: float
        Standard deviation of the distribution (:math:`\sigma_\chi`)

    Returns
    -------
    array-like: The probability
    """
    return truncnorm(dataset["chi_p"], mu=mu_chi_p, sigma=sigma_chi_p, low=0, high=1)


class GaussianChiEffChiP(object):
    r"""
    A covariant Gaussian in effective aligned and precessing spins.

    .. math::

        p(\chi_{\rm eff}, \chi_p | \Lambda) = \mathcal{N}_{[-1, 1], [0, 1]}(\chi_{\rm eff}, \chi_p; [\mu_{\rm eff}, \mu_{p}], \Sigma)

    Where :math:`\mathcal{N}_{[a, b], [c, d]}` is the two-dimensional truncated normal distribution
    over :math:`[a, b]` and :math:`[c, d]`.

    The covariance matrix is given by:

    .. math::
        \Sigma = \begin{bmatrix}
            \sigma^2_{\text{eff}} & \rho \sigma_{\text{eff}} \sigma_{p} \\
            \rho \sigma_{\text{eff}} \sigma_{p} & \sigma^2_{p}
        \end{bmatrix}

    See `Miller+ <https://arxiv.org/abs/2001.06051>`_ and `Callister+ <https://arxiv.org/abs/2010.14533>`_.

    Parameters
    ----------
    dataset: dict
        Dictionary of numpy arrays for :code:`chi_eff` and :code:`chi_p`.
    mu_chi_eff: float
        Mean of the chi effective distribution (:math:`\mu_{\text{eff}}`)
    mu_chi_p: float
        Mean of the chi p distribution (:math:`\mu_{p}`)
    sigma_chi_eff: float
        Standard deviation of the chi effective distribution (:math:`\sigma_{\text{eff}}`)
    sigma_chi_p: float
        Standard deviation of the chi p distribution (:math:`\sigma_{p}`)
    spin_covariance: float
        Covariance between the two parameters (:math:`\rho`)
    """

    def __init__(self):
        self.chi_eff = xp.linspace(-1, 1, 500)
        self.chi_p = xp.linspace(0, 1, 250)
        self.chi_eff_grid, self.chi_p_grid = xp.meshgrid(self.chi_eff, self.chi_p)

    def __call__(
        self, dataset, mu_chi_eff, sigma_chi_eff, mu_chi_p, sigma_chi_p, spin_covariance
    ):

        prob = unnormalized_2d_gaussian(
            dataset["chi_eff"],
            dataset["chi_p"],
            mu_chi_eff,
            mu_chi_p,
            sigma_chi_eff,
            sigma_chi_p,
            spin_covariance,
        )
        normalization = self._normalization(
            mu_chi_eff=mu_chi_eff,
            sigma_chi_eff=sigma_chi_eff,
            mu_chi_p=mu_chi_p,
            sigma_chi_p=sigma_chi_p,
            spin_covariance=spin_covariance,
        )
        prob /= normalization
        prob *= xp.abs(dataset["chi_eff"]) <= 1
        prob *= (dataset["chi_p"] <= 1) * (dataset["chi_p"] >= 0)
        return prob

    def _normalization(
        self, mu_chi_eff, sigma_chi_eff, mu_chi_p, sigma_chi_p, spin_covariance
    ):
        r"""
        Numerically calculate the normalization over a two-dimensional grid with
        trapezoidal integration

        Parameters
        ----------
        mu_chi_eff: float
            Mean of the chi effective distribution (:math:`\mu_{\text{eff}}`)
        mu_chi_p: float
            Mean of the chi p distribution (:math:`\mu_{p}`)
        sigma_chi_eff: float
            Standard deviation of the chi effective distribution (:math:`\sigma_{\text{eff}}`)
        sigma_chi_p: float
            Standard deviation of the chi p distribution (:math:`\sigma_{p}`)
        spin_covariance: float
            Covariance between the two parameters (:math:`\rho`)

        Returns
        -------
        float
            The normalizing constant
        """
        prob = unnormalized_2d_gaussian(
            self.chi_eff_grid,
            self.chi_p_grid,
            mu_chi_eff,
            mu_chi_p,
            sigma_chi_eff,
            sigma_chi_p,
            spin_covariance,
        )
        return trapezoid(
            y=trapezoid(y=prob, axis=-1, x=self.chi_eff), axis=-1, x=self.chi_p
        )


class SplineSpinMagnitudeIdentical(InterpolatedNoBaseModelIdentical):
    """
    Interpolated spline model for spin magnitudes.

    See `Golomb and Talbot <https://arxiv.org/abs/2210.12287>`_

    Parameters
    ----------
    minimum: float
        Minimum value to normalize the spline over, default=0.
    maximum: float
        Maximum value to normalize the spline over, default=1.
    nodes: int
        Number of nodes to use in the spline, default=5.
    kind: str
        The interpolation order of the spline, default=”cubic”.
    regularize: bool
        Whether to regularize the spline node values to have root-mean-square
        value :code:`rms{name}`, default=False.
    """

    def __init__(self, minimum=0, maximum=1, nodes=5, kind="cubic", regularize=False):

        super(SplineSpinMagnitudeIdentical, self).__init__(
            parameters=["a_1", "a_2"],
            minimum=minimum,
            maximum=maximum,
            nodes=nodes,
            kind=kind,
            regularize=regularize,
        )


class SplineSpinTiltIdentical(InterpolatedNoBaseModelIdentical):
    """
    Interpolated spline model for spin orientations.

    See `Golomb and Talbot <https://arxiv.org/abs/2210.12287>`_

    Parameters
    ----------
    minimum: float
        Minimum value to normalize the spline over, default=-1.
    maximum: float
        Maximum value to normalize the spline over, default=1.
    nodes: int
        Number of nodes to use in the spline, default=5.
    kind: str
        The interpolation order of the spline, default=”cubic”.
    regularize: bool
        Whether to regularize the spline node values to have root-mean-square
        value :code:`rms{name}`, default=False.
    """

    def __init__(self, minimum=-1, maximum=1, nodes=5, kind="cubic", regularize=False):

        super(SplineSpinTiltIdentical, self).__init__(
            parameters=["cos_tilt_1", "cos_tilt_2"],
            minimum=minimum,
            maximum=maximum,
            nodes=nodes,
            kind=kind,
            regularize=regularize,
        )



from gwpopulation.conversions import *

known_values = dict(
    alpha=[1, 2, 2],
    beta=[1, 4, 4],
    amax=[1, 1, 0.5],
    mu=[1 / 2, 1 / 3, 1 / 6],
    var=[1 / 12, 2 / 63, 1 / 126],
)
suffices = ["", "_1", "_2"]


def test_mu_chi_var_chi_max_to_alpha_beta_max():
    for ii in range(3):
        mu = known_values["mu"][ii]
        var = known_values["var"][ii]
        amax = known_values["amax"][ii]
        alpha, beta, _ = mu_var_max_to_alpha_beta_max(mu, var, amax)
        assert abs(alpha - known_values["alpha"][ii]) < 1e-6
        assert abs(beta == known_values["beta"][ii]) < 1e-6
        assert amax == known_values["amax"][ii]


def test_alpha_beta_max_to_mu_chi_var_chi_max():
    for ii in range(3):
        alpha = known_values["alpha"][ii]
        beta = known_values["beta"][ii]
        amax = known_values["amax"][ii]
        mu, var, _ = alpha_beta_max_to_mu_var_max(alpha, beta, amax)
        assert mu == known_values["mu"][ii]
        assert var == known_values["var"][ii]


def test_convert_to_beta_parameters():
    for ii, suffix in enumerate(suffices):
        params = dict()
        params["mu_chi" + suffix] = known_values["mu"][ii]
        params["sigma_chi" + suffix] = known_values["var"][ii]
        params["amax" + suffix] = known_values["amax"][ii]
        new_params, _ = convert_to_beta_parameters(params, remove=True)
        full_dict = params.copy()
        alpha, beta, _ = mu_var_max_to_alpha_beta_max(
            mu=params["mu_chi" + suffix],
            var=params["sigma_chi" + suffix],
            amax=params["amax" + suffix],
        )
        full_dict["alpha_chi" + suffix] = alpha
        full_dict["beta_chi" + suffix] = beta
        print(new_params, full_dict)
        for key in full_dict:
            assert new_params[key] == full_dict[key]


def test_convert_to_beta_parameters_with_none():
    params = dict(amax=1, alpha_chi=None, beta_chi=None, mu_chi=0.5, sigma_chi=0.1)
    _, added = convert_to_beta_parameters(params, remove=True)
    assert len(added) == 2


def test_convert_to_beta_parameters_unnecessary():
    params = dict(amax=1, alpha_chi=1, beta_chi=1)
    _, added = convert_to_beta_parameters(params, remove=True)
    assert len(added) == 0



"""
Cosmological functionality in :code:`GWPopulation` is based on the :code:`wcosmo` package.
For more details see the `wcosmo documentation <https://wcosmo.readthedocs.io/en/latest/>`_.

We provide a mixin class :func:`gwpopulation.experimental.cosmo_models.CosmoMixin` that
can be used to add cosmological functionality to a population model.
"""

import numpy as xp
from bilby.hyper.model import Model
from wcosmo import z_at_value
from wcosmo.astropy import WCosmoMixin, available
from wcosmo.utils import disable_units as wcosmo_disable_units


class CosmoMixin:
    """
    Mixin class that provides cosmological functionality to a subclass.

    Parameters
    ==========
    cosmo_model: str
        The cosmology model to use. Default is :code:`Planck15`.
        Should be of :code:`wcosmo.available.keys()`.
    """

    def __init__(self, cosmo_model="Planck15"):
        wcosmo_disable_units()
        self.cosmo_model = cosmo_model
        if self.cosmo_model == "FlatwCDM":
            self.cosmology_names = ["H0", "Om0", "w0"]
        elif self.cosmo_model == "FlatLambdaCDM":
            self.cosmology_names = ["H0", "Om0"]
        else:
            self.cosmology_names = []
        self._cosmo = available[cosmo_model]

    def cosmology_variables(self, parameters):
        """
        Extract the cosmological parameters from the provided parameters.

        Parameters
        ==========
        parameters: dict
            The parameters for the cosmology model.

        Returns
        =======
        dict
            A dictionary containing :code:`self.cosmology_names` as keys.
        """
        return {key: parameters[key] for key in self.cosmology_names}

    def cosmology(self, parameters):
        """
        Return the cosmology model given the parameters.

        Parameters
        ==========
        parameters: dict
            The parameters for the cosmology model.

        Returns
        =======
        wcosmo.astropy.WCosmoMixin
            The cosmology model.
        """
        if isinstance(self._cosmo, WCosmoMixin):
            return self._cosmo
        else:
            return self._cosmo(**self.cosmology_variables(parameters))

    def detector_frame_to_source_frame(self, data, **parameters):
        r"""
        Convert detector frame samples to source frame samples given cosmological
        parameters. Calculate the corresponding
        :math:`\frac{d \theta_{\rm detector}}{d \theta_{\rm source}}` Jacobian term.
        This includes factors of :math:`(1 + z)` for redshifted quantities.

        Parameters
        ==========
        data: dict
            Dictionary containing the samples in detector frame.
        parameters: dict
            The cosmological parameters for relevant cosmology model.

        Returns
        =======
        samples: dict
            Dictionary containing the samples in source frame.
        jacobian: array-like
            The Jacobian term.
        """

        samples = dict()
        if "luminosity_distance" in data.keys():
            cosmo = self.cosmology(parameters)
            samples["redshift"] = z_at_value(
                cosmo.luminosity_distance,
                data["luminosity_distance"],
            )
            jacobian = cosmo.dDLdz(samples["redshift"])
        elif "redshift" not in data:
            raise ValueError(
                "Either luminosity_distance or redshift must be provided in data for detector frame to source frame samples conversion"
            )
        else:
            jacobian = xp.ones(data["redshift"].shape)

        for key in list(data.keys()):
            if key.endswith("_detector"):
                samples[key[:-9]] = data[key] / (1 + samples["redshift"])
                jacobian *= 1 + samples["redshift"]
            elif key != "luminosity_distance":
                samples[key] = data[key]

        return samples, jacobian


class CosmoModel(Model, CosmoMixin):
    """
    Modified version of :code:`bilby.hyper.model.Model` that automatically
    updates the source-frame quantities given the detector-frame quantities and
    cosmology and disables caching due to the source-frame quantities changing
    every iteration.

    Parameters
    ==========
    model_functions: list
        List containing the model functions.
    cosmo_model: str
        The cosmology model to use. Default is :code:`Planck15`.
        Should be of :code:`wcosmo.available.keys()`.
    """

    def __init__(self, model_functions=None, cosmo_model="Planck15"):
        Model.__init__(self, model_functions=model_functions, cache=False)
        CosmoMixin.__init__(self, cosmo_model=cosmo_model)

    def prob(self, data, **kwargs):
        """
        Compute the total population probability for the provided data given
        the keyword arguments.

        This method augments :code:`bilby.hyper.model.Model.prob` by converting
        the detector frame samples to source frame samples and dividing by the
        corresponding Jacobian term.

        Parameters
        ==========
        data: dict
            Dictionary containing the points at which to evaluate the
            population model.
        kwargs: dict
            The population parameters. These cannot include any of
            :code:`["dataset", "data", "self", "cls"]` unless the
            :code:`variable_names` attribute is available for the relevant
            model.
        """

        data, jacobian = self.detector_frame_to_source_frame(data, **kwargs)
        probability = super().prob(data, **kwargs)
        probability /= jacobian

        return probability



"""
:code:`GWPopulation` is a Python package for doing population inference
with the gravitational-wave transient catalogue supporting a range of
numpy-like backends.

This includes:

- commonly used likelihood functions in the Bilby framework.
- population models for gravitational-wave sources.
- selection functions for gravitational-wave sources.
- spectral siren cosmological inference with flat wCDM models.
"""

from . import conversions, hyperpe, models, utils, vt
from ._version import __version__
from .backend import SUPPORTED_BACKENDS, set_backend



import importlib

import pytest

import gwpopulation


def test_unsupported_backend_raises_value_error():
    with pytest.raises(ValueError):
        gwpopulation.set_backend("fail")


def test_set_backend(backend):
    gwpopulation.set_backend(backend)
    from gwpopulation.utils import xp

    assert xp == importlib.import_module(gwpopulation.backend._np_module[backend])


def test_import_error_caught_for_mangled_install():
    """
    Replace importlib.import_module with a dummy function raise
    the required error.

    Two calls are needed to avoid caching of the backend.

    FIXME: figure out how to replace this with mock
    """

    def _import(module):
        raise ImportError

    gwpopulation.backend.import_module = _import
    with pytest.raises(ImportError):
        gwpopulation.set_backend("numpy")
        gwpopulation.set_backend("jax")

    gwpopulation.backend.import_module = importlib.import_module


def test_loading_arbitrary():
    """
    Test loading arbitrary functions works as we don't have any native
    entry points for them.
    """
    pytest.importorskip("jax")

    from jax.scipy.linalg import toeplitz

    func = gwpopulation.backend._load_arbitrary(
        func="scipy.linalg.toeplitz", backend="jax"
    )
    assert func == toeplitz



import numpy as np
import pytest
from bilby.hyper.model import Model

import gwpopulation
from gwpopulation import vt
from gwpopulation.models.redshift import PowerLawRedshift, total_four_volume

xp = np


def dummy_function(dataset, alpha):
    return 1


def test_initialize_with_function():
    evaluator = vt._BaseVT(model=dummy_function, data=dict())
    assert evaluator.model.models == [dummy_function]


def test_initialize_with_list_of_functions():
    evaluator = vt._BaseVT(model=[dummy_function, dummy_function], data=dict())
    assert evaluator.model.models == [dummy_function, dummy_function]


def test_initialize_with_bilby_model():
    model = Model([dummy_function, dummy_function])
    evaluator = vt._BaseVT(model=model, data=dict())
    assert evaluator.model.models == [dummy_function, dummy_function]


def test_base_cannot_be_called():
    model = Model([dummy_function, dummy_function])
    evaluator = vt._BaseVT(model=model, data=dict())
    with pytest.raises(NotImplementedError):
        evaluator()


def test_grid_vt_correct(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = lambda dataset: xp.ones_like(dataset["a"])
    data = dict(a=xp.linspace(0, 1, 1000), vt=2)
    assert float(vt.GridVT(model, data)(dict())) == 2


def resampling_data(xp):
    return dict(a=xp.linspace(0, 1, 1000), prior=xp.ones(1000), vt=2)


def get_vt(xp):
    data = resampling_data(xp)
    model = lambda dataset: (
        xp.exp(-((dataset["a"] - 0.5) ** 2) / 2) / (2 * xp.pi) ** 0.5
    )
    evaluator = vt.ResamplingVT(data=data, model=model, n_events=0)
    return evaluator


def test_marginalized_vt_correct(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    assert abs(float(get_vt(xp).vt_factor(dict())) - 0.38289403358409585) < 1e-6


def test_resampling_vt_correct(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    assert abs(float(get_vt(xp)(dict())[0]) - 0.38289325179141254) < 1e-6


def test_returns_inf_when_n_effective_too_small(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    evaluator = get_vt(xp)
    evaluator.marginalize_uncertainty = True
    evaluator.n_events = xp.inf
    assert evaluator(dict()) == xp.inf


def test_observed_volume_with_no_redshift_model(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    evaluator = get_vt(xp)
    assert evaluator.surveyed_hypervolume(dict()) == total_four_volume(
        lamb=0, analysis_time=1
    )


def test_observed_volume_with_redshift_model(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    data = resampling_data(xp)
    model = PowerLawRedshift()
    evaluator = vt.ResamplingVT(data=data, model=model, n_events=0)
    assert (
        abs(
            evaluator.surveyed_hypervolume(dict(lamb=4))
            - total_four_volume(lamb=4, analysis_time=1)
        )
        < 1e-4
    )



from importlib import import_module

__backend__ = ""
SUPPORTED_BACKENDS = ["numpy", "cupy", "jax"]
_np_module = dict(numpy="numpy", cupy="cupy", jax="jax.numpy")
_scipy_module = dict(numpy="scipy", cupy="cupyx.scipy", jax="jax.scipy")

__all__ = [
    "SUPPORTED_BACKENDS",
    "modules_to_update",
    "set_backend",
]

__doc__ = f"""
:code:`GWPopulation` provides a unified interface to a number of :code:`numpy/scipy` like APIs.

The backend can be set using :code:`gwpopulation.set_backend(backend)`, where
:code:`backend` is one of :code:`{', '.join(SUPPORTED_BACKENDS)}`.

Downstream packages can automatically track the active backend using :code:`entry_points`.
With this set up, packages can use :code:`xp` and :code:`scs` in specified modules.
Additionally, users can provide a full arbitrary scipy object to be used if anything beyond
:code:`scipy.special` is needed.
An example of how to set :code:`numpy`, :code:`scipy.special`, and the :code:`toeplitz` function
from :code:`scipy.linalg` via the :code:`setup.cfg` file is shown below.
Specification using :code:`pyproject.toml` and :code:`setup.py` follows slightly
different syntax documentation for which can be found online.

.. code-block::

    [options.entry_points]
    gwpopulation.xp =
        mypackage_foo = mypackage.foo
    gwpopulation.scs =
        mypackage_foo = mypackage.foo
        mypackage_bar = mypackage.bar
    gwpopulation.other =
        mypackage_baz_toeplitz = mypackage.baz:scipy.linalg.toeplitz

.. note::
    Each module that wants to use the :code:`GWPopulation` backend must be specified independently
    for the automatic propagation to work.

If there is a backend that you would like to use that is not currently supported, please open an issue.
"""


def modules_to_update():
    """
    Return all modules that need to be updated with the backend.

    Returns
    -------
    all_with_xp: list
        List of all modules that need to be updated with the backend's :code:`xp`
        (:code:`numpy`).
    all_with_scs: list
        List of all modules that need to be updated with the backend's :code:`scs`
        (:code:`scipy.special`).
    others: dict
        Dictionary of all modules that need to be updated with arbitrary functions
        from :code:`scipy`.
    """
    import sys

    if sys.version_info < (3, 10):
        from importlib_metadata import entry_points
    else:
        from importlib.metadata import entry_points
    all_with_xp = [module.value for module in entry_points(group="gwpopulation.xp")]
    all_with_scs = [module.value for module in entry_points(group="gwpopulation.scs")]
    other_entries = [
        module.value.split(":") for module in entry_points(group="gwpopulation.other")
    ]
    others = {key: value for key, value in other_entries}
    return all_with_xp, all_with_scs, others


def _configure_jax(xp):
    """
    Configuration requirements for :code:`jax`

    - use 64-bit floats.
    """
    from jax import config

    config.update("jax_enable_x64", True)


def _load_numpy_and_scipy(backend):
    try:
        xp = import_module(_np_module[backend])
        scs = import_module(_scipy_module[backend]).special
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f"{backend} not installed for gwpopulation")
    except ImportError:
        raise ImportError(f"{backend} installed but not importable for gwpopulation")

    if backend == "jax":
        _configure_jax(xp)

    return xp, scs


def _load_arbitrary(func, backend):
    if func.startswith("scipy"):
        func = func.replace("scipy", _scipy_module[backend])
    elif func.startswith("numpy"):
        func = func.replace("numpy", _np_module[backend])
    module, func = func.rsplit(".", 1)
    return getattr(import_module(module), func)


def set_backend(backend="numpy"):
    """
    Set the backend for :code:`GWPopulation` and plugins.

    This will automatically update all modules that have been registered to use
    the :code:`GWPopulation` automatic backend tracking.

    .. warning::

        This will not update existing instances of classes.

    Parameters
    ----------
    backend: str
        The backend to use, one of the :code:`SUPPORTED_BACKENDS`.

    Raises
    ------
    ValueError
        If the backend is not in :code:`SUPPORTED_BACKENDS`.
    """
    global __backend__
    if backend not in SUPPORTED_BACKENDS:
        raise ValueError(
            f"Backend {backend} not supported, should be in {', '.join(SUPPORTED_BACKENDS)}"
        )
    elif backend == __backend__:
        return

    xp, scs = _load_numpy_and_scipy(backend)

    __backend__ = backend
    all_with_xp, all_with_scs, others = modules_to_update()
    for module in all_with_xp:
        setattr(import_module(module), "xp", xp)
    for module in all_with_scs:
        setattr(import_module(module), "scs", scs)
    for module, func in others.items():
        setattr(
            import_module(module),
            func.split(".")[-1],
            _load_arbitrary(func, backend),
        )



"""Tests for gwpopulation.experimental.cosmo_models module."""

import numpy as np
import pytest
from wcosmo.utils import disable_units

from gwpopulation.experimental.cosmo_models import CosmoMixin, CosmoModel


class TestCosmoMixin:
    """Test suite for CosmoMixin class."""

    def test_init_planck15(self):
        """Test initialization with Planck15 (fixed cosmology)."""
        mixin = CosmoMixin(cosmo_model="Planck15")
        assert mixin.cosmo_model == "Planck15"
        assert mixin.cosmology_names == []

    def test_init_flatwcdm(self):
        """Test initialization with FlatwCDM (parametrized cosmology)."""
        mixin = CosmoMixin(cosmo_model="FlatwCDM")
        assert mixin.cosmo_model == "FlatwCDM"
        assert mixin.cosmology_names == ["H0", "Om0", "w0"]

    def test_init_flatlambdacdm(self):
        """Test initialization with FlatLambdaCDM (parametrized cosmology)."""
        mixin = CosmoMixin(cosmo_model="FlatLambdaCDM")
        assert mixin.cosmo_model == "FlatLambdaCDM"
        assert mixin.cosmology_names == ["H0", "Om0"]

    def test_cosmology_variables_flatwcdm(self):
        """Test extracting cosmology variables for FlatwCDM."""
        mixin = CosmoMixin(cosmo_model="FlatwCDM")
        parameters = {"H0": 70, "Om0": 0.3, "w0": -1.0, "other_param": 5}
        cosmo_vars = mixin.cosmology_variables(parameters)
        assert cosmo_vars == {"H0": 70, "Om0": 0.3, "w0": -1.0}
        assert "other_param" not in cosmo_vars

    def test_cosmology_variables_flatlambdacdm(self):
        """Test extracting cosmology variables for FlatLambdaCDM."""
        mixin = CosmoMixin(cosmo_model="FlatLambdaCDM")
        parameters = {"H0": 67.4, "Om0": 0.315, "other_param": 5}
        cosmo_vars = mixin.cosmology_variables(parameters)
        assert cosmo_vars == {"H0": 67.4, "Om0": 0.315}
        assert "other_param" not in cosmo_vars

    def test_cosmology_planck15(self):
        """Test cosmology method with Planck15 (no parameters needed)."""
        disable_units()
        mixin = CosmoMixin(cosmo_model="Planck15")
        parameters = {}
        cosmo = mixin.cosmology(parameters)
        # For Planck15, should return the same instance
        assert cosmo is not None
        # Verify it's a valid cosmology object with expected properties
        assert hasattr(cosmo, "luminosity_distance")
        assert hasattr(cosmo, "dDLdz")

    def test_cosmology_flatwcdm(self):
        """Test cosmology method with FlatwCDM (parametrized)."""
        disable_units()
        mixin = CosmoMixin(cosmo_model="FlatwCDM")
        parameters = {"H0": 70, "Om0": 0.3, "w0": -1.0}
        cosmo = mixin.cosmology(parameters)
        assert cosmo is not None
        assert hasattr(cosmo, "luminosity_distance")
        assert hasattr(cosmo, "dDLdz")

    def test_cosmology_flatlambdacdm(self):
        """Test cosmology method with FlatLambdaCDM (parametrized)."""
        disable_units()
        mixin = CosmoMixin(cosmo_model="FlatLambdaCDM")
        parameters = {"H0": 67.4, "Om0": 0.315}
        cosmo = mixin.cosmology(parameters)
        assert cosmo is not None
        assert hasattr(cosmo, "luminosity_distance")
        assert hasattr(cosmo, "dDLdz")

    def test_detector_frame_to_source_frame_with_redshift(self):
        """Test detector_frame_to_source_frame when redshift is provided."""
        disable_units()
        mixin = CosmoMixin(cosmo_model="Planck15")
        data = {
            "redshift": np.array([0.1, 0.2, 0.3]),
            "mass_1_detector": np.array([30.0, 40.0, 50.0]),
            "other_param": np.array([1.0, 2.0, 3.0]),
        }
        parameters = {}

        samples, jacobian = mixin.detector_frame_to_source_frame(data, **parameters)

        # Check that redshift is preserved
        np.testing.assert_array_equal(samples["redshift"], data["redshift"])

        # Check that detector frame mass is converted to source frame
        expected_mass_1 = data["mass_1_detector"] / (1 + data["redshift"])
        np.testing.assert_allclose(samples["mass_1"], expected_mass_1)

        # Check that other parameters are preserved
        np.testing.assert_array_equal(samples["other_param"], data["other_param"])

        # Check jacobian for detector frame quantities
        expected_jacobian = 1 + data["redshift"]
        np.testing.assert_allclose(jacobian, expected_jacobian)

    def test_detector_frame_to_source_frame_with_luminosity_distance_planck15(self):
        """Test detector_frame_to_source_frame with luminosity distance (Planck15)."""
        disable_units()
        mixin = CosmoMixin(cosmo_model="Planck15")
        # Use realistic luminosity distances
        data = {
            "luminosity_distance": np.array([400.0, 800.0, 1200.0]),
        }
        parameters = {}

        samples, jacobian = mixin.detector_frame_to_source_frame(data, **parameters)

        # Check that redshift is computed
        assert "redshift" in samples
        assert len(samples["redshift"]) == 3
        # Redshift should be positive for positive luminosity distances
        assert np.all(samples["redshift"] > 0)

        # Check that jacobian is computed
        assert len(jacobian) == 3
        assert np.all(jacobian > 0)

    def test_detector_frame_to_source_frame_with_luminosity_distance_flatwcdm(self):
        """Test detector_frame_to_source_frame with FlatwCDM and luminosity distance."""
        disable_units()
        mixin = CosmoMixin(cosmo_model="FlatwCDM")
        data = {
            "luminosity_distance": np.array([400.0, 800.0, 1200.0]),
            "mass_1_detector": np.array([30.0, 40.0, 50.0]),
        }
        # This is the critical test: parameters must be passed through
        parameters = {"H0": 70, "Om0": 0.3, "w0": -1.0}

        samples, jacobian = mixin.detector_frame_to_source_frame(data, **parameters)

        # Check that redshift is computed
        assert "redshift" in samples
        assert len(samples["redshift"]) == 3
        assert np.all(samples["redshift"] > 0)

        # Check that detector frame mass is converted
        expected_mass_1 = data["mass_1_detector"] / (1 + samples["redshift"])
        np.testing.assert_allclose(samples["mass_1"], expected_mass_1, rtol=1e-10)

        # Check that jacobian includes both dDLdz and (1+z) terms
        assert len(jacobian) == 3
        assert np.all(jacobian > 0)

    def test_detector_frame_to_source_frame_with_luminosity_distance_flatlambdacdm(
        self,
    ):
        """Test detector_frame_to_source_frame with FlatLambdaCDM and luminosity distance."""
        disable_units()
        mixin = CosmoMixin(cosmo_model="FlatLambdaCDM")
        data = {
            "luminosity_distance": np.array([400.0, 800.0, 1200.0]),
        }
        # This is the critical test: parameters must be passed through
        parameters = {"H0": 67.4, "Om0": 0.315}

        samples, jacobian = mixin.detector_frame_to_source_frame(data, **parameters)

        # Check that redshift is computed
        assert "redshift" in samples
        assert len(samples["redshift"]) == 3
        assert np.all(samples["redshift"] > 0)

        # Check that jacobian is computed
        assert len(jacobian) == 3
        assert np.all(jacobian > 0)

    def test_detector_frame_to_source_frame_missing_redshift_and_luminosity_distance(
        self,
    ):
        """Test that ValueError is raised when neither redshift nor luminosity distance is provided."""
        disable_units()
        mixin = CosmoMixin(cosmo_model="Planck15")
        data = {"mass_1_detector": np.array([30.0, 40.0, 50.0])}
        parameters = {}

        with pytest.raises(ValueError):
            mixin.detector_frame_to_source_frame(data, **parameters)

    def test_detector_frame_to_source_frame_preserves_non_detector_keys(self):
        """Test that non-detector frame keys are preserved correctly."""
        disable_units()
        mixin = CosmoMixin(cosmo_model="Planck15")
        data = {
            "redshift": np.array([0.1, 0.2, 0.3]),
            "mass_1_detector": np.array([30.0, 40.0, 50.0]),
            "mass_2_detector": np.array([20.0, 25.0, 30.0]),
            "spin_1": np.array([0.1, 0.2, 0.3]),
            "luminosity_distance": np.array([400.0, 800.0, 1200.0]),
        }
        parameters = {}

        samples, jacobian = mixin.detector_frame_to_source_frame(data, **parameters)

        # Check that redshift is preserved
        np.testing.assert_array_equal(samples["redshift"], data["redshift"])

        # Check that detector frame quantities are converted
        assert "mass_1" in samples
        assert "mass_2" in samples

        # Check that non-detector quantities are preserved
        np.testing.assert_array_equal(samples["spin_1"], data["spin_1"])

        # Check that luminosity_distance is not in samples (it's not copied)
        assert "luminosity_distance" not in samples

    def test_detector_frame_to_source_frame_multiple_detector_quantities(self):
        """Test conversion with multiple detector frame quantities."""
        disable_units()
        mixin = CosmoMixin(cosmo_model="Planck15")
        data = {
            "redshift": np.array([0.1, 0.2, 0.3]),
            "mass_1_detector": np.array([30.0, 40.0, 50.0]),
            "mass_2_detector": np.array([20.0, 25.0, 30.0]),
            "chirp_mass_detector": np.array([25.0, 30.0, 35.0]),
        }
        parameters = {}

        samples, jacobian = mixin.detector_frame_to_source_frame(data, **parameters)

        # Check all detector frame quantities are converted
        for key in ["mass_1", "mass_2", "chirp_mass"]:
            assert key in samples
            expected = data[key + "_detector"] / (1 + data["redshift"])
            np.testing.assert_allclose(samples[key], expected)

        # Jacobian should be (1+z)^3 for three detector frame quantities
        expected_jacobian = (1 + data["redshift"]) ** 3
        np.testing.assert_allclose(jacobian, expected_jacobian)

    def test_detector_frame_to_source_frame_parameters_different_cosmologies(self):
        """Test that different parameters lead to different redshifts for same luminosity distance."""
        disable_units()
        mixin = CosmoMixin(cosmo_model="FlatLambdaCDM")
        data = {
            "luminosity_distance": np.array([1000.0]),
        }

        # Test with different H0 values
        parameters1 = {"H0": 60.0, "Om0": 0.3}
        parameters2 = {"H0": 80.0, "Om0": 0.3}

        samples1, _ = mixin.detector_frame_to_source_frame(data, **parameters1)
        samples2, _ = mixin.detector_frame_to_source_frame(data, **parameters2)

        # Different H0 should lead to different redshifts
        assert samples1["redshift"][0] != samples2["redshift"][0]


class TestCosmoModel:
    """Test suite for CosmoModel class."""

    def test_init_default(self):
        """Test initialization with default parameters."""
        disable_units()
        model = CosmoModel(model_functions=[])
        assert model.cosmo_model == "Planck15"
        assert model.cosmology_names == []
        assert model.models == []

    def test_init_with_cosmo_model(self):
        """Test initialization with custom cosmology model."""
        disable_units()
        model = CosmoModel(model_functions=[], cosmo_model="FlatwCDM")
        assert model.cosmo_model == "FlatwCDM"
        assert model.cosmology_names == ["H0", "Om0", "w0"]

    def test_init_with_model_functions(self):
        """Test initialization with model functions."""
        disable_units()

        def simple_model(dataset, **parameters):
            """Simple test model that returns constant probability."""
            return np.ones(len(dataset.get("mass_1", [1])))

        model = CosmoModel(model_functions=[simple_model])
        assert len(model.models) == 1

    def test_prob_with_planck15_and_redshift(self):
        """Test prob method with Planck15 cosmology and redshift provided."""
        disable_units()

        def simple_model(dataset, **parameters):
            """Simple model that depends on mass_1."""
            return np.ones(len(dataset["mass_1"]))

        model = CosmoModel(model_functions=[simple_model], cosmo_model="Planck15")

        data = {
            "redshift": np.array([0.1, 0.2, 0.3]),
            "mass_1_detector": np.array([30.0, 40.0, 50.0]),
        }
        parameters = {}

        prob = model.prob(data, **parameters)

        # Check that probability is computed and jacobian is applied
        assert len(prob) == 3
        assert np.all(prob > 0)

        # Expected jacobian is (1 + z) for one detector frame quantity
        expected_jacobian = 1 + data["redshift"]
        # Probability should be 1 / jacobian since simple_model returns ones
        np.testing.assert_allclose(prob, 1.0 / expected_jacobian)

    def test_prob_with_planck15_and_luminosity_distance(self):
        """Test prob method with Planck15 cosmology and luminosity distance."""
        disable_units()

        def simple_model(dataset, **parameters):
            """Simple model that depends on mass_1."""
            return np.ones(len(dataset["mass_1"]))

        model = CosmoModel(model_functions=[simple_model], cosmo_model="Planck15")

        data = {
            "luminosity_distance": np.array([400.0, 800.0, 1200.0]),
            "mass_1_detector": np.array([30.0, 40.0, 50.0]),
        }
        parameters = {}

        prob = model.prob(data, **parameters)

        # Check that probability is computed
        assert len(prob) == 3
        assert np.all(prob > 0)

    def test_prob_with_flatwcdm_passes_kwargs(self):
        """Test that prob method passes kwargs to detector_frame_to_source_frame for FlatwCDM."""
        disable_units()

        def simple_model(dataset, **parameters):
            """Simple model that returns ones."""
            return np.ones(len(dataset["mass_1"]))

        model = CosmoModel(model_functions=[simple_model], cosmo_model="FlatwCDM")

        data = {
            "luminosity_distance": np.array([1000.0]),
            "mass_1_detector": np.array([30.0]),
        }
        # These parameters MUST be passed through for FlatwCDM to work
        parameters = {"H0": 70.0, "Om0": 0.3, "w0": -1.0}

        # This should NOT raise an error - the bug would cause it to fail
        # because kwargs were not passed to detector_frame_to_source_frame
        prob = model.prob(data, **parameters)

        assert len(prob) == 1
        assert prob[0] > 0

    def test_prob_with_flatlambdacdm_passes_kwargs(self):
        """Test that prob method passes kwargs to detector_frame_to_source_frame for FlatLambdaCDM."""
        disable_units()

        def simple_model(dataset, **parameters):
            """Simple model that returns ones."""
            return np.ones(len(dataset["mass_1"]))

        model = CosmoModel(model_functions=[simple_model], cosmo_model="FlatLambdaCDM")

        data = {
            "luminosity_distance": np.array([1000.0]),
            "mass_1_detector": np.array([30.0]),
        }
        # These parameters MUST be passed through for FlatLambdaCDM to work
        parameters = {"H0": 67.4, "Om0": 0.315}

        # This should NOT raise an error
        prob = model.prob(data, **parameters)

        assert len(prob) == 1
        assert prob[0] > 0

    def test_prob_with_flatwcdm_different_parameters(self):
        """Test that different cosmological parameters lead to different probabilities."""
        disable_units()

        def simple_model(dataset, **parameters):
            """Simple model that returns ones."""
            return np.ones(len(dataset["mass_1"]))

        model = CosmoModel(model_functions=[simple_model], cosmo_model="FlatwCDM")

        data = {
            "luminosity_distance": np.array([1000.0]),
            "mass_1_detector": np.array([30.0]),
        }

        # Test with different H0 values
        parameters1 = {"H0": 60.0, "Om0": 0.3, "w0": -1.0}
        parameters2 = {"H0": 80.0, "Om0": 0.3, "w0": -1.0}

        prob1 = model.prob(data, **parameters1)
        prob2 = model.prob(data, **parameters2)

        # Different cosmological parameters should lead to different probabilities
        assert prob1[0] != prob2[0]

    def test_prob_applies_jacobian_correctly(self):
        """Test that the jacobian is applied correctly for multiple detector frame quantities."""
        disable_units()

        def simple_model(dataset, **parameters):
            """Simple model that returns ones."""
            return np.ones(len(dataset["mass_1"]))

        model = CosmoModel(model_functions=[simple_model], cosmo_model="Planck15")

        data = {
            "redshift": np.array([0.1, 0.2, 0.3]),
            "mass_1_detector": np.array([30.0, 40.0, 50.0]),
            "mass_2_detector": np.array([20.0, 25.0, 30.0]),
        }
        parameters = {}

        prob = model.prob(data, **parameters)

        # Expected jacobian is (1 + z)^2 for two detector frame quantities
        expected_jacobian = (1 + data["redshift"]) ** 2
        # Probability should be 1 / jacobian since simple_model returns ones
        np.testing.assert_allclose(prob, 1.0 / expected_jacobian)

    def test_prob_with_complex_model(self):
        """Test prob method with a more complex model function."""
        disable_units()

        def mass_dependent_model(dataset, alpha=2.0, **parameters):
            """Model that depends on mass_1 with a power law."""
            return dataset["mass_1"] ** (-alpha)

        model = CosmoModel(
            model_functions=[mass_dependent_model], cosmo_model="Planck15"
        )

        data = {
            "redshift": np.array([0.1, 0.2]),
            "mass_1_detector": np.array([30.0, 40.0]),
        }
        parameters = {"alpha": 2.35}

        prob = model.prob(data, **parameters)

        # Check that probability is computed correctly
        assert len(prob) == 2
        assert np.all(prob > 0)

        # Verify the calculation:
        # mass_1 = mass_1_detector / (1 + z)
        mass_1_source = data["mass_1_detector"] / (1 + data["redshift"])
        # Model returns mass_1^(-alpha)
        model_prob = mass_1_source ** (-2.35)
        # Jacobian is (1 + z) for one detector frame quantity
        jacobian = 1 + data["redshift"]
        expected_prob = model_prob / jacobian

        np.testing.assert_allclose(prob, expected_prob)



import glob

import bilby
import pandas as pd
import pytest
from bilby.hyper.model import Model

import gwpopulation
from gwpopulation.experimental.jax import JittedLikelihood


@pytest.mark.parametrize("jit", [True, False])
def test_likelihood_evaluation(backend, jit):
    if backend != "jax" and jit:
        pytest.skip(reason="JIT only works with JAX")

    gwpopulation.set_backend(backend)
    xp = gwpopulation.models.mass.xp
    bilby.core.utils.random.seed(10)
    rng = bilby.core.utils.random.rng

    if jit:
        cache = False
    else:
        cache = True

    model = Model(
        [
            gwpopulation.models.mass.SinglePeakSmoothedMassDistribution(),
            gwpopulation.models.spin.independent_spin_magnitude_beta,
            gwpopulation.models.spin.independent_spin_orientation_gaussian_isotropic,
            gwpopulation.models.redshift.PowerLawRedshift(),
        ],
        cache=cache,
    )
    vt_model = Model(
        [
            gwpopulation.models.mass.SinglePeakSmoothedMassDistribution(),
            gwpopulation.models.spin.independent_spin_magnitude_beta,
            gwpopulation.models.spin.independent_spin_orientation_gaussian_isotropic,
            gwpopulation.models.redshift.PowerLawRedshift(),
        ],
        cache=cache,
    )

    bounds = dict(
        mass_1=(20, 25),
        mass_ratio=(0.9, 1),
        a_1=(0, 1),
        a_2=(0, 1),
        cos_tilt_1=(-1, -1),
        cos_tilt_2=(-1, -1),
        redshift=(0, 2),
        prior=(1, 1),
    )
    posteriors = [
        pd.DataFrame({key: rng.uniform(*bound, 100) for key, bound in bounds.items()})
        for _ in range(10)
    ]
    vt_data = {
        key: xp.asarray(rng.uniform(*bound, 10000)) for key, bound in bounds.items()
    }

    selection = gwpopulation.vt.ResamplingVT(vt_model, vt_data, len(posteriors))

    likelihood = gwpopulation.hyperpe.HyperparameterLikelihood(
        hyper_prior=model,
        posteriors=posteriors,
        selection_function=selection,
    )

    priors = bilby.core.prior.PriorDict("priors/bbh_population.prior")
    prior_sample = priors.sample()
    assert abs(likelihood.log_likelihood_ratio(prior_sample) + 7.037596674351107) < 0.01
    if jit:
        likelihood = JittedLikelihood(likelihood)
    assert abs(likelihood.log_likelihood_ratio(prior_sample) + 7.037596674351107) < 0.01
    likelihood.posterior_predictive_resample(pd.DataFrame(priors.sample(5)))


def test_prior_files_load():
    for fname in glob.glob("priors/*.prior"):
        print(fname)
        _ = bilby.core.prior.PriorDict(fname)



import itertools

import numpy as np
import pytest
from bilby.core.prior import DeltaFunction, Normal, PriorDict, Uniform

import gwpopulation
from gwpopulation.models import mass
from gwpopulation.utils import to_numpy, trapezoid

xp = np
N_TEST = 10


def double_power_prior():
    power_prior = PriorDict()
    power_prior["alpha_1"] = Uniform(minimum=-4, maximum=12)
    power_prior["alpha_2"] = Uniform(minimum=-4, maximum=12)
    power_prior["beta"] = Uniform(minimum=-4, maximum=12)
    power_prior["mmin"] = Uniform(minimum=3, maximum=10)
    power_prior["mmax"] = Uniform(minimum=40, maximum=100)
    power_prior["break_fraction"] = Uniform(minimum=0, maximum=1)
    return power_prior


def power_prior():
    power_prior = PriorDict()
    power_prior["alpha"] = Uniform(minimum=-4, maximum=12)
    power_prior["beta"] = Uniform(minimum=-4, maximum=12)
    power_prior["mmin"] = Uniform(minimum=3, maximum=10)
    power_prior["mmax"] = Uniform(minimum=40, maximum=100)
    return power_prior


def gauss_prior():
    gauss_prior = PriorDict()
    gauss_prior["lam"] = Uniform(minimum=0, maximum=1)
    gauss_prior["mpp"] = Uniform(minimum=20, maximum=60)
    gauss_prior["sigpp"] = Uniform(minimum=0, maximum=10)
    return gauss_prior


def double_gauss_prior():
    double_gauss_prior = PriorDict()
    double_gauss_prior["lam"] = Uniform(minimum=0, maximum=1)
    double_gauss_prior["lam_1"] = Uniform(minimum=0, maximum=1)
    double_gauss_prior["mpp_1"] = Uniform(minimum=20, maximum=60)
    double_gauss_prior["mpp_2"] = Uniform(minimum=20, maximum=100)
    double_gauss_prior["sigpp_1"] = Uniform(minimum=0, maximum=10)
    double_gauss_prior["sigpp_2"] = Uniform(minimum=0, maximum=10)
    return double_gauss_prior


def smooth_prior():
    smooth_prior = PriorDict()
    smooth_prior["delta_m"] = Uniform(minimum=0, maximum=10)
    return smooth_prior


def interpolated_prior(n_nodes):
    prior = PriorDict()
    nodes = np.logspace(np.log10(2), np.log10(100), n_nodes)
    for ii in range(n_nodes):
        prior[f"mass{ii}"] = DeltaFunction(nodes[ii])
        prior[f"fmass{ii}"] = Normal(0, 1)
    return prior


def test_double_power_law_zero_below_mmin(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    to_numpy = gwpopulation.utils.to_numpy
    m1s, _, _ = get_primary_mass_ratio_data(xp)
    prior = double_power_prior()
    for ii in range(N_TEST):
        parameters = prior.sample()
        del parameters["beta"]
        p_m = mass.double_power_law_primary_mass(m1s, **parameters)
        p_m = to_numpy(p_m)
        assert np.max(p_m[to_numpy(m1s) < parameters["mmin"]]) == 0.0


def test_power_law_primary_mass_ratio_zero_above_mmax(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    to_numpy = gwpopulation.utils.to_numpy
    m1s, qs, dataset = get_primary_mass_ratio_data(xp)
    prior = double_power_prior()
    m1s = dataset["mass_1"]
    for ii in range(N_TEST):
        parameters = prior.sample()
        p_m = mass.double_power_law_primary_power_law_mass_ratio(dataset, **parameters)
        p_m = to_numpy(p_m)
        assert np.max(p_m[to_numpy(m1s) > parameters["mmax"]]) == 0.0


def get_primary_mass_ratio_data(xp):
    m1s = xp.linspace(3, 100, 1000)
    qs = xp.linspace(0.01, 1, 500)
    m1s_grid, qs_grid = xp.meshgrid(m1s, qs)
    dataset = dict(mass_1=m1s_grid, mass_ratio=qs_grid)
    dataset["mass_2"] = dataset["mass_1"] * dataset["mass_ratio"]
    return m1s, qs, dataset


def get_primary_secondary_data(xp):
    ms = xp.linspace(3, 100, 1000)
    dm = ms[1] - ms[0]
    m1s_grid, m2s_grid = xp.meshgrid(ms, ms)
    dataset = dict(mass_1=m1s_grid, mass_2=m2s_grid)
    return to_numpy(ms), float(dm), dataset


def get_smoothed_data(xp):
    m1s = xp.linspace(2, 100, 1000)
    qs = xp.linspace(0.01, 1, 500)
    m1s_grid, qs_grid = xp.meshgrid(m1s, qs)
    dataset = dict(mass_1=m1s_grid, mass_ratio=qs_grid)
    return m1s, qs, dataset


analytic_models = {
    "single": mass.power_law_primary_mass_ratio,
    "double": mass.double_power_law_primary_power_law_mass_ratio,
    "identical": mass.power_law_primary_secondary_identical,
    "independent": mass.power_law_primary_secondary_independent,
    "two": mass.two_component_primary_mass_ratio,
    "two-identical": mass.two_component_primary_secondary_identical,
    "two-independent": mass.two_component_primary_secondary_independent,
}


def analytic_prior(model):
    if model in ["single", "independent"]:
        return power_prior()
    elif model == "double":
        return double_power_prior()
    elif model == "identical":
        prior = power_prior()
        del prior["beta"]
        return prior
    elif model in ["two", "two-independent"]:
        prior = power_prior()
        prior.update(gauss_prior())
        prior["gaussian_mass_maximum"] = 80
        return prior
    elif model == "two-identical":
        prior = power_prior()
        prior.update(gauss_prior())
        prior["gaussian_mass_maximum"] = 80
        del prior["beta"]
        return prior
    raise ValueError(f"{model} not known")


@pytest.mark.parametrize("model", analytic_models.keys())
def test_zero_below_mmin(backend, model):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    to_numpy = gwpopulation.utils.to_numpy
    _, _, dataset = get_primary_mass_ratio_data(xp)
    prior = analytic_prior(model)
    model = analytic_models[model]
    m2s = dataset["mass_2"]
    for _ in range(N_TEST):
        parameters = prior.sample()
        p_m = to_numpy(model(dataset, **parameters))
        assert np.max(p_m[to_numpy(m2s) < parameters["mmin"]]) == 0.0


@pytest.mark.parametrize("model", analytic_models.keys())
def test_zero_above_mmax(backend, model):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    to_numpy = gwpopulation.utils.to_numpy
    _, _, dataset = get_primary_mass_ratio_data(xp)
    prior = analytic_prior(model)
    model = analytic_models[model]
    m1s = dataset["mass_1"]
    for _ in range(N_TEST):
        parameters = prior.sample()
        p_m = to_numpy(model(dataset, **parameters))
        mmax = max(parameters["mmax"], parameters.get("gaussian_mass_maximum", 0))
        assert np.max(p_m[to_numpy(m1s) > mmax]) == 0.0


def test_single_peak_delta_m_zero_matches_two_component_primary_mass_ratio(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    _, _, dataset = get_smoothed_data(xp)
    max_diffs = list()
    prior = power_prior()
    prior.update(gauss_prior())
    model = mass.SinglePeakSmoothedMassDistribution()
    for _ in range(N_TEST):
        parameters = prior.sample()
        p_m1 = mass.two_component_primary_mass_ratio(dataset, **parameters)
        parameters["delta_m"] = 0
        p_m2 = model(dataset, **parameters)
        max_diffs.append(_max_abs_difference(p_m1, p_m2, xp=xp))
    assert max(max_diffs) < 1e-5


def test_double_peak_delta_m_zero_matches_three_component_primary_mass_ratio(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    _, _, dataset = get_smoothed_data(xp)
    max_diffs = list()
    prior = power_prior()
    prior.update(double_gauss_prior())
    model = mass.MultiPeakSmoothedMassDistribution()
    for _ in range(N_TEST):
        parameters = prior.sample()
        del parameters["beta"]
        p_m1 = mass.three_component_single(mass=dataset["mass_1"], **parameters)
        parameters["delta_m"] = 0
        p_m2 = model.p_m1(dataset, **parameters)
        max_diffs.append(_max_abs_difference(p_m1, p_m2, xp=xp))
    assert max(max_diffs) < 1e-5


def _normalised(model, prior, xp):
    """
    Add jit to jax testing as these functions involve
    caching and so we burn a call and then jit.
    """
    m1s, qs, dataset = get_smoothed_data(xp)
    parameters = prior.sample()
    _ = model(dataset, **parameters)

    if "jax" in xp.__name__:
        import jax

        model = jax.jit(model)

    norms = list()
    for _ in range(N_TEST):
        parameters = prior.sample()
        p_m = model(dataset, **parameters)
    norms.append(float(trapezoid(trapezoid(p_m, m1s), qs)))
    assert _max_abs_difference(norms, 1.0, xp=xp) < 0.01


def test_single_peak_normalised(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = mass.SinglePeakSmoothedMassDistribution()
    prior = power_prior()
    prior.update(gauss_prior())
    prior.update(smooth_prior())
    _normalised(model, prior, xp)


def test_double_peak_normalised(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = mass.MultiPeakSmoothedMassDistribution()
    prior = power_prior()
    prior.update(double_gauss_prior())
    prior.update(smooth_prior())
    _normalised(model, prior, xp)


def test_broken_power_law_normalised(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = mass.BrokenPowerLawSmoothedMassDistribution()
    prior = double_power_prior()
    prior.update(smooth_prior())
    _normalised(model, prior, xp)


def test_broken_power_law_peak_normalised(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = mass.BrokenPowerLawPeakSmoothedMassDistribution()
    prior = double_power_prior()
    prior.update(smooth_prior())
    prior.update(gauss_prior())
    _normalised(model, prior, xp)


def test_set_minimum_and_maximum(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = mass.SinglePeakSmoothedMassDistribution(mmin=5, mmax=150)
    parameters = gauss_prior().sample()
    parameters.update(power_prior().sample())
    parameters.update(smooth_prior().sample())
    parameters["mpp"] = 130
    parameters["sigpp"] = 1
    parameters["lam"] = 0.5
    parameters["mmin"] = 6
    assert (
        model(dict(mass_1=8 * xp.ones(6), mass_ratio=0.5 * xp.ones(6)), **parameters)[0]
        == 0
    )
    assert (
        model(dict(mass_1=130 * xp.ones(5), mass_ratio=0.9 * xp.ones(5)), **parameters)[
            0
        ]
        > 0
    )


def test_mmin_below_global_minimum_raises_error():
    gwpopulation.set_backend("numpy")
    model = mass.SinglePeakSmoothedMassDistribution(mmin=5, mmax=150)
    parameters = gauss_prior().sample()
    parameters.update(power_prior().sample())
    parameters.update(smooth_prior().sample())
    parameters["mmin"] = 2
    with pytest.raises(ValueError):
        model(dict(mass_1=5, mass_ratio=0.9), **parameters)


def test_mmax_above_global_maximum_raises_error():
    gwpopulation.set_backend("numpy")
    model = mass.SinglePeakSmoothedMassDistribution(mmin=5, mmax=150)
    parameters = gauss_prior().sample()
    parameters.update(power_prior().sample())
    parameters.update(smooth_prior().sample())
    parameters["mmax"] = 200
    with pytest.raises(ValueError):
        model(dict(mass_1=5, mass_ratio=0.9), **parameters)


def test_interpolated_power_law_p_m1_normalised(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = mass.InterpolatedPowerlaw()
    prior = interpolated_prior(10)
    prior.update(smooth_prior())
    prior.update(power_prior())
    _normalised(model, prior, xp)


def test_interpolated_power_law_p_m1_variable_names():
    model = mass.InterpolatedPowerlaw()
    prior = interpolated_prior(10)
    prior.update(smooth_prior())
    prior.update(power_prior())
    assert model.variable_names == set(prior.keys())


def _max_abs_difference(array, comparison, xp=np):
    return float(xp.max(xp.abs(comparison - xp.asarray(array))))



import jax
import jax.numpy as jnp
import numpy as np
from numpyro import deterministic, sample
from numpyro.distributions import Gamma, Normal, Unit

__all__ = [
    "gwpopulation_likelihood_model",
    "posterior_predictive_resample",
]


def gwpopulation_likelihood_model(
    likelihood,
    hyper_params,
    varmax=np.inf,
    apply_selection=True,
    predictive_resample=True,
    fit_keys=None,
):
    """
    A :code:`numpyro` implementation of :func:`gwpopulation.hyperpe.HyperparameterLikelihood`.

    Parameters
    ==========
    likelihood: gwpopulation.hyperpe.HyperparameterLikelihood
        The likelihood to evaluate.
    hyper_params: dict
        The population parameters to use.
    varmax: float, optional
        The maximum allowed variance in the likelihood.
    apply_selection: bool, optional
        Whether to apply selection effects. Default is :code:`True`.
    predictive_resample: bool, optional
        Whether to draw samples from the predicitve distribution.
        Default is :code:`True`.
    fit_keys: list, optional
        The keys to resample. Default is :code:`None`, which fits all keys in :code:data.
    """
    parameters, added_keys = likelihood.conversion_function(hyper_params.copy())

    weights = (
        likelihood.hyper_prior.prob(likelihood.data, **parameters)
        / likelihood.sampling_prior
    )
    expectations = jnp.mean(weights, axis=-1)
    square_expectations = jnp.mean(weights**2, axis=-1)
    variances = deterministic(
        "variances",
        (square_expectations - expectations**2)
        / (likelihood.samples_per_posterior * expectations**2),
    )
    ln_bayes_factors = deterministic("ln_bayes_factors", jnp.log(expectations))

    ln_l_events = jnp.sum(ln_bayes_factors)
    variance_events = jnp.sum(variances)

    if apply_selection:
        selector = likelihood.selection_function
        injection_weights = (
            selector.model.prob(selector.data, **parameters) / selector.data["prior"]
        )
        efficiency = jnp.sum(injection_weights) / selector.total_injections
        selection_uncertainty = (
            jnp.sum(injection_weights**2) / selector.total_injections**2
            - efficiency**2 / selector.total_injections
        )
        selection = deterministic(
            "selection", -jnp.log(efficiency) * likelihood.n_posteriors
        )
        selection_variance = deterministic(
            "selection_variance",
            selection_uncertainty / efficiency**2 * likelihood.n_posteriors**2,
        )

        total_hypervolume = selector.surveyed_hypervolume(parameters)
        vt = total_hypervolume * efficiency
        _ = sample("rate", Gamma(likelihood.n_posteriors, 1 / vt))
        if predictive_resample:
            posterior_predictive_resample(
                selector.data,
                injection_weights,
                "injections",
                shape=(likelihood.n_posteriors,),
                fit_keys=fit_keys,
            )
    else:
        selection = 0
        selection_variance = 0

    ln_l = deterministic("ln_l", ln_l_events + selection)
    variance = deterministic("variance", variance_events + selection_variance)

    if predictive_resample:
        for event in range(likelihood.n_posteriors):
            posterior_predictive_resample(
                likelihood.data, weights, event, event=event, fit_keys=fit_keys
            )
    if varmax < np.inf:
        ln_l = jnp.where(
            jnp.isnan(ln_l) | (variance > varmax),
            jnp.nan_to_num(-jnp.inf),
            ln_l,
        )
    return ln_l


def posterior_predictive_resample(
    data, weights, label, event=None, shape=(), fit_keys=None
):
    """
    Draw a sample from the predictive distribution given the weights for each sample.

    Parameters
    ----------
    data: dict
        The data dictionary.
    weights: array-like
        The weights for each sample.
    label: str
        The label to use for the new sample.
    event: int, optional
        The event number to use. Default is :code:`None`, assumes the input is
        one dimensional.
    shape: tuple, optional
        The shape of the output sample. Default is :code:`()`.
    fit_keys: list, optional
        The keys to fit. Default is :code:`None`, which fits all keys in :code:data.
    """
    if event is not None:
        weights = weights[:, event]
    weights /= jnp.sum(weights)
    n_points = int(weights.shape[0])
    idx = jax.random.choice(
        jax.random.PRNGKey(0),
        jnp.arange(n_points),
        p=weights,
        shape=shape,
        replace=False,
    )
    if fit_keys is None:
        fit_keys = data.keys()
    for key in fit_keys:
        if event is None:
            deterministic(f"{key}_{label}", data[key][idx])
        else:
            deterministic(f"{key}_{label}", data[key][event, idx])



import numpy as np


# coding: utf-8
def circle(radius, phi, size=400, color="black"):
    center = size / 2
    positions = (
        center + radius * np.cos(phi),
        center + radius * np.sin(phi),
        center - radius * np.cos(phi),
        center - radius * np.sin(phi),
        center + radius * np.cos(phi - np.pi / 6),
        center + radius * np.sin(phi - np.pi / 6),
        center - radius * np.cos(phi - np.pi / 6),
        center - radius * np.sin(phi - np.pi / 6),
    )
    width = 2 * radius / 100
    inner = radius / 10
    ends = (
        center + radius * np.cos(phi + np.arctan(1 / 10)),
        center + radius * np.sin(phi + np.arctan(1 / 10)),
        center - radius * np.cos(phi + np.arctan(1 / 10)),
        center - radius * np.sin(phi + np.arctan(1 / 10)),
    )
    return f"""
  <!-- <circle cx="{center}" cy="{center}" r="{radius}" stroke="{color}" stroke-width="2" fill="none" /> -->
  <circle cx="{positions[0]}" cy="{positions[1]}" r="{inner}" stroke="{color}" stroke-width="{width}" fill="none" />
  <circle cx="{positions[2]}" cy="{positions[3]}" r="{inner}" stroke="{color}" stroke-width="{width}" fill="none" />

  <path d="M {ends[0]} {ends[1]} A {radius} {radius} 0 0 1 {positions[6]} {positions[7]}" stroke="{color}" stroke-width="{width}" fill="none" />
  <path d="M {ends[2]} {ends[3]} A {radius} {radius} 0 0 1 {positions[4]} {positions[5]}" stroke="{color}" stroke-width="{width}" fill="none" />
    """


def logo(size=400, color="rgb(40,10,40)", center="rgb(240,240,240)", text=False):

    gradient = f"""  <!-- Define the radial gradient -->
  <defs>
    <radialGradient id="grad1" cx="50%" cy="50%" r="50%" fx="50%" fy="50%">
      <stop offset="00%" style="stop-color:{center};stop-opacity:1" />
      <stop offset="50%" style="stop-color:{center};stop-opacity:1" />
      <stop offset="90%" style="stop-color:{color};stop-opacity:1" />
    </radialGradient>
  </defs>"""
    data = f"""<svg width="{size * 5**int(text)}" height="{size}" xmlns="http://www.w3.org/2000/svg">"""
    #     data += """
    #   {gradient}

    #   <!-- Use the gradient for the background -->
    #   <!-- <rect x="0" y="0" width="100%" height="100%" fill="url(#grad1)" /> -->
    # """
    # data += circle(120, 0.3 * np.pi, size=size)
    largest = size / 4
    data += circle(largest, -0.1 * np.pi, size=size, color=color)
    data += circle(largest / 5, 0.65 * np.pi, size=size, color=color)
    data += circle(largest * 2 / 5, 0.15 * np.pi, size=size, color=color)
    data += circle(largest * 3 / 5, 0.4 * np.pi, size=size, color=color)
    data += circle(largest * 4 / 5, -0.3 * np.pi, size=size, color=color)

    if text:
        data += f"""\t<text x="{size * 0.9}" y="{size * 0.68}" font-family="Verdana" font-size="{size * 0.5}" fill="{color}">GWPopulation</text>"""
    data += "\n</svg>"
    return data


if __name__ == "__main__":
    print(
        logo(size=40, color="rgb(240,240,240)", center="rgb(215,245,215)", text=False),
        file=open("_static/logo-dark.svg", "w"),
    )
    print(
        logo(size=40, color="rgb(15,15,15)", center="rgb(215,245,215)", text=False),
        file=open("_static/logo.svg", "w"),
    )
    print(
        logo(size=400, color="rgb(240,240,240)", center="rgb(215,245,215)", text=True),
        file=open("_static/logo-long-dark.svg", "w"),
    )
    print(
        logo(size=400, color="rgb(15,15,15)", center="rgb(215,245,215)", text=True),
        file=open("_static/logo-long.svg", "w"),
    )



from functools import partial

import numpy as np

from ..utils import to_numpy, trapezoid

xp = np

__all__ = [
    "InterpolatedNoBaseModelIdentical",
    "_setup_interpolant",
]


def _setup_interpolant(nodes, values, kind="cubic", backend=None):
    """
    Create a caching spline interpolant.

    .. note::
        This function is not intended to be used directly but can overwrite
        the :code:`GWPopulation` backend.

    Parameters
    ==========
    nodes: array-like
        The positions of the nodes
    values: array-like
        The positions that the interpolant will be evaluated at.
    kind: str
        The interpolation order of the spline, default="cubic"
    backend: module
        The backend to use, default=:code:`numpy`.
    """
    from cached_interpolate import RegularCachingInterpolant as CachingInterpolant

    if backend is None:
        backend = xp

    nodes = to_numpy(nodes)
    interpolant = CachingInterpolant(nodes, nodes, kind=kind, backend=backend)
    interpolant.conversion = backend.array(interpolant.conversion)
    interpolant = partial(interpolant, backend.array(values))
    return interpolant


class InterpolatedNoBaseModelIdentical:
    """
    Base class for the Interpolated classes with no base model

    Parameters
    ==========
    parameters: list
        List of parameters to interpolate over, e.g., :code:`["a_1", "a_2"]`
    minimum: float
        Minimum value to normalize the spline over
    maximum: float
        Maximum value to normalize the spline over
    nodes: int
        Number of nodes to use in the spline, default=10
    kind: str
        The interpolation order of the spline, default=:code:`"cubic"`
    log_nodes: bool
        Whether to use log-spaced nodes, default=False
    regularize: bool
        Whether to regularize the spline node values to have root-mean-square value
        :code:`rms{name}`, default=False
    """

    def __init__(
        self,
        parameters,
        minimum,
        maximum,
        nodes=10,
        kind="cubic",
        log_nodes=False,
        regularize=False,
    ):
        self.nodes = nodes
        self._norm_spline = None
        self._data_spline = dict()
        self.kind = kind
        self._xs = xp.linspace(minimum, maximum, 10 * self.nodes)
        self.parameters = parameters
        self.min = minimum
        self.max = maximum
        self.log_nodes = log_nodes

        self.base = self.parameters[0].strip("_1")
        self.xkeys = [f"{self.base}{ii}" for ii in range(self.nodes)]
        self.fkeys = [f"f{self.base}{ii}" for ii in range(self.nodes)]
        self.regularize = regularize

    def __call__(self, dataset, **kwargs):
        """
        A wrapper to :func:`p_x_identical`
        """
        return self.p_x_identical(dataset, **kwargs)

    @property
    def variable_names(self):
        """
        The names of the hyperparameters of the model.

        These are the names of the parameters describing:

        - the node locations
        - the value of the model at the nodes
        - the root-mean-square value of the model if :code:`self.regularize=True`
        """
        keys = self.xkeys + self.fkeys
        if self.regularize:
            keys += [f"rms{self.base}"]
        return keys

    def setup_interpolant(self, nodes, values):
        if self.log_nodes:
            func = xp.log
        else:
            func = xp.array
        kwargs = dict(kind=self.kind, backend=xp)
        self._norm_spline = _setup_interpolant(func(nodes), func(self._xs), **kwargs)
        self._data_spline = {
            param: _setup_interpolant(func(nodes), func(values[param]), **kwargs)
            for param in self.parameters
        }

    def p_x_unnormed(self, dataset, parameter, x_splines, f_splines, **kwargs):
        """
        Calculate the unnormalized likelihood of the dataset given the model

        Parameters
        ==========
        dataset: dict
            Dictionary containing the data to evaluate the likelihood at
        parameter: str
            The parameter to evaluate the likelihood for
        x_splines: array-like
            The positions of the spline nodes
        f_splines: array-like
            The values at the spline nodes
        kwargs: dict, UNUSED

        Returns
        =======
        p_x: array-like
            The unnormalized likelihood of the dataset given the model at the
            provided points
        """
        if self._norm_spline is None:
            self.setup_interpolant(x_splines, dataset)

        perturbation = self._data_spline[parameter](y=f_splines)

        p_x = xp.exp(perturbation)
        p_x *= (dataset[f"{parameter}"] >= x_splines[0]) & (
            dataset[f"{parameter}"] <= x_splines[-1]
        )
        return p_x

    def norm_p_x(self, f_splines=None, x_splines=None, **kwargs):
        """
        Calculate the normalization of the spline

        Parameters
        ==========
        f_splines: array-like
            The values at the spline nodes
        x_splines: array-like
            The positions of the spline nodes
        kwargs: dict, UNUSED

        Returns
        =======
        norm: float
            The normalization of the spline
        """
        perturbation = self._norm_spline(y=f_splines)
        p_x = xp.exp(perturbation)
        p_x *= (self._xs >= x_splines[0]) & (self._xs <= x_splines[-1])
        norm = trapezoid(p_x, self._xs)
        return norm

    def extract_spline_points(self, kwargs):
        """
        Extract the node positions and values from the dictionary of parameters

        Parameters
        ==========
        kwargs: dict
            Dictionary containing :code:`{self.base}_ii, f{self.base_ii}` and
            optionally :code`rms{self.base}`

        Returns
        =======
        f_splines: array-like
            The values at the spline nodes
        x_splines: array-like
            The positions of the spline nodes
        """
        f_splines = xp.array([kwargs[key] for key in self.fkeys])
        if self.regularize:
            f_splines *= kwargs[f"rms{self.base}"] / xp.mean(f_splines**2) ** 0.5
        x_splines = xp.array([kwargs[key] for key in self.xkeys])
        return f_splines, x_splines

    def p_x_identical(self, dataset, **kwargs):
        """
        Calculate the likelihood of the dataset given the model assuming
        that all the parameters are identically distributed.

        Parameters
        ==========
        dataset: dict
            Dictionary containing the data to evaluate the likelihood at
        kwargs: dict
            Dictionary containing the parameters of the model

        Returns
        =======
        p_x: array-like
            The likelihood of the dataset given the model at the provided points
        """
        self.infer_n_nodes(**kwargs)

        f_splines, x_splines = self.extract_spline_points(kwargs)

        p_x = xp.ones(xp.shape(dataset[self.parameters[0]]))

        for param in self.parameters:
            p_x *= self.p_x_unnormed(
                dataset, param, x_splines=x_splines, f_splines=f_splines, **kwargs
            )

        norm = self.norm_p_x(f_splines=f_splines, x_splines=x_splines, **kwargs)
        p_x /= norm ** len(self.parameters)
        return p_x

    def infer_n_nodes(self, **kwargs):
        """
        Infer the number of nodes from the dictionary of parameters.
        This method looks for the first missing parameter matching
        :code:`f{self.base}{nodes}` and updates the number of nodes
        to the number of found parameters.
        """
        nodes = self.nodes

        while True:
            if f"f{self.base}{nodes}" in kwargs:
                nodes += 1
            else:
                break

        if not nodes == self.nodes:
            print(
                f"Different number of nodes! Using {nodes} nodes instead of {self.nodes}"
            )
            self.__init__(nodes=nodes)



"""
Helper functions for probability distributions and backend switching.
"""

from numbers import Number
from operator import ge, gt, ne

import numpy as np
from scipy import special as scs

xp = np


def apply_conditions(conditions):
    """
    A decorator to apply conditions to inputs of a function.

    Parameters
    ==========
    func: callable
        The function to decorate.
    conditions: dict
        A dictionary of conditions to apply to the function. The keys are the
        argument names and the values are the conditions. The conditions can be
        should be in the form (op, value) where op is a comparison operator and
        value is the value to compare to. The conditions can also be a callable
        which takes the value as an argument and returns a boolean. The variable
        must be specified as a keyword argument for the test to be applied.
    """
    from functools import wraps

    def decorator(func):
        @wraps(func)
        def wrapped_function(*args, **kwargs):
            if "jax" in xp.__name__:
                return func(*args, **kwargs)
            for key, condition in conditions.items():
                if key in kwargs:
                    value = kwargs[key]
                else:
                    continue
                if callable(condition):
                    if not condition(value):
                        raise ValueError(f"Condition {condition} not met")
                else:
                    op, val = condition
                    if "cupy" in xp.__name__:
                        value = xp.asarray(value)
                    if callable(op):
                        if not xp.all(op(value, val)):
                            raise ValueError(
                                f"{key}: {value} does not satisfy {op.__name__}"
                            )
                    else:
                        raise ValueError(f"Operator {op} not supported")
            return func(*args, **kwargs)

        return wrapped_function

    return decorator


@apply_conditions(dict(alpha=(gt, 0), beta=(gt, 0), scale=(gt, 0)))
def beta_dist(xx, alpha, beta, scale=1):
    r"""
    Beta distribution probability

    .. math::
        p(x) = \frac{x^{\alpha - 1} (x_\max - x)^{\beta - 1}}{B(\alpha, \beta) x_\max^{\alpha + \beta + 1}}

    Parameters
    ----------
    xx: float, array-like
        The abscissa values (:math:`x`)
    alpha: float
        The Beta alpha parameter (:math:`\alpha`)
    beta: float
        The Beta beta parameter (:math:`\beta`)
    scale: float, array-like
        A scale factor for the distribution of the distribution (:math:`x_\max`)

    Returns
    -------
    prob: float, array-like
        The distribution evaluated at `xx`

    """
    ln_beta = (alpha - 1) * xp.log(xx) + (beta - 1) * xp.log(scale - xx)
    ln_beta -= scs.betaln(alpha, beta)
    ln_beta -= (alpha + beta - 1) * xp.log(scale)
    prob = xp.exp(ln_beta)
    prob = xp.nan_to_num(prob)
    prob *= (xx >= 0) * (xx <= scale)
    return prob


@apply_conditions(dict(low=(ge, 0), alpha=(ne, 1)))
def powerlaw(xx, alpha, high, low):
    r"""
    Power-law probability

    .. math::
        p(x) = \frac{1 + \alpha}{x_\max^{1 + \alpha} - x_\min^{1 + \alpha}} x^\alpha

    Parameters
    ----------
    xx: float, array-like
        The abscissa values (:math:`x`)
    alpha: float, array-like
        The spectral index of the distribution (:math:`\alpha`)
    high: float, array-like
        The maximum of the distribution (:math:`x_\min`)
    low: float, array-like
        The minimum of the distribution (:math:`x_\max`)

    Returns
    -------
    prob: float, array-like
        The distribution evaluated at `xx`

    """
    norm = xp.where(
        xp.array(alpha) == -1,
        1 / xp.log(high / low),
        (1 + alpha) / xp.array(high ** (1 + alpha) - low ** (1 + alpha)),
    )
    prob = xp.power(xx, alpha)
    prob *= norm
    prob *= (xx <= high) & (xx >= low)
    return prob


@apply_conditions(dict(sigma=(gt, 0)))
def truncnorm(xx, mu, sigma, high, low):
    r"""
    Truncated normal probability

    .. math::

        p(x) =
        \sqrt{\frac{2}{\pi\sigma^2}}\frac{\exp\left(-\frac{(\mu - x)^2}{2 \sigma^2}\right)}
        {\text{erf}\left(\frac{x_\max - \mu}{\sqrt{2}}\right) + \text{erf}\left(\frac{\mu - x_\min}{\sqrt{2}}\right)}

    Parameters
    ----------
    xx: float, array-like
        The abscissa values (:math:`x`)
    mu: float, array-like
        The mean of the normal distribution (:math:`\mu`)
    sigma: float
        The standard deviation of the distribution (:math:`\sigma`)
    high: float, array-like
        The maximum of the distribution (:math:`x_\min`)
    low: float, array-like
        The minimum of the distribution (:math:`x_\max`)

    Returns
    -------
    prob: float, array-like
        The distribution evaluated at `xx`

    """

    def logsubexp(log_p, log_q):
        return log_p + xp.log(1 - xp.exp(log_q - log_p))

    zz = xp.array(xx - mu) / sigma
    aa = xp.array(low - mu) / sigma
    bb = xp.array(high - mu) / sigma
    log_pdf = -(zz**2) / 2.0 - np.log(2.0 * np.pi) / 2.0 - xp.log(sigma)

    # cf https://github.com/scipy/scipy/blob/v1.15.1/scipy/stats/_continuous_distns.py#L10189
    log_norm = xp.select(
        [bb <= 0, aa > 0, bb > 0],
        [
            logsubexp(scs.log_ndtr(bb), scs.log_ndtr(aa)),
            logsubexp(scs.log_ndtr(-aa), scs.log_ndtr(-bb)),
            xp.log1p(-scs.ndtr(aa) - scs.ndtr(-bb)),
        ],
        xp.nan,
    )
    log_pdf -= log_norm
    return xp.nan_to_num(xp.exp(log_pdf)) * (xx >= low) * (xx <= high)


def unnormalized_2d_gaussian(xx, yy, mu_x, mu_y, sigma_x, sigma_y, covariance):
    r"""
    Compute the probability distribution for a correlated 2-dimensional Gaussian
    neglecting normalization terms.

    .. math::
        \ln p(x) &= (x - \mu_{x} y - \mu_{y}) \Sigma^{-1} (x - \mu_x y - \mu_{y})^{T} \\
        \Sigma &= \begin{bmatrix}
                \sigma^{2}_{x} & \rho \sigma_{x} \sigma_{y} \\
                \rho \sigma_{x} \sigma_{y} & \sigma^{2}_{y}
            \end{bmatrix}

    Parameters
    ----------
    xx: array-like
        Input data in first dimension (:math:`x`)
    yy: array-like
        Input data in second dimension (:math:`y`)
    mu_x: float
        Mean in the first dimension (:math:`\mu_{x}`)
    sigma_x: float
        Standard deviation in the first dimension (:math:`\sigma_{x}`)
    mu_y: float
        Mean in the second dimension (:math:`\mu_{y}`)
    sigma_y: float
        Standard deviation in the second dimension (:math:`\sigma_{y}`)
    covariance: float
        The normalized covariance (bounded in [0, 1]) (:math:`\rho`)

    Returns
    -------
    prob: array-like
        The unnormalized probability distribution (:math:`p(x)`)
    """
    determinant = sigma_x**2 * sigma_y**2 * (1 - covariance)
    residual_x = (mu_x - xx) * sigma_y
    residual_y = (mu_y - yy) * sigma_x
    prob = xp.exp(
        -(residual_x**2 + residual_y**2 - 2 * residual_x * residual_y * covariance)
        / 2
        / determinant
    )
    return prob


def von_mises(xx, mu, kappa):
    r"""
    PDF of the von Mises distribution defined on the standard interval.

    .. math::
        p(x) =
        \frac{\exp\left( \kappa \cos(x - \mu) \right)}{2 \pi \mathcal{I}_{0}(\kappa)}

    Parameters
    ----------
    xx: array-like
        Input points at which to evaluate the distribution
    mu: float
        The mean of the distribution
    kappa: float
        The scale parameter

    Returns
    -------
    array-like
        The probability

    Notes
    -----
    For numerical stability, the factor of :math:`\exp(\kappa)` from using
    :func:`scipy.special.i0e` is accounted for in the numerator.
    """
    return xp.exp(kappa * (xp.cos(xx - mu) - 1)) / (2 * xp.pi * scs.i0e(kappa))


def get_name(input):
    """
    Attempt to find the name of the the input. This either returns
    :code:`input.__name__` or :code:`input.__class__.__name__`

    Parameters
    ==========
    input: any
        The input to find the name for.

    Returns
    =======
    str: The name of the input.
    """
    if hasattr(input, "__name__"):
        return input.__name__
    else:
        return input.__class__.__name__


def to_number(value, func):
    """
    Convert a zero-dimensional array to a number.

    Parameters
    ==========
    value: array-like
        The zero-dimensional array to convert.
    func: callable
        The function to convert the array with,
        e.g., :code:`int,float,complex`.
    """
    if "jax" in xp.__name__:
        return value.astype(func)
    else:
        return func(value)


def to_numpy(array):
    """
    Convert an array to a numpy array.
    Numeric types and pandas objects are returned unchanged.

    Parameters
    ==========
    array: array-like
        The array to convert.
    """
    if isinstance(array, (Number, np.ndarray)):
        return array
    elif "cupy" in array.__class__.__module__:
        return xp.asnumpy(array)
    elif "pandas" in array.__class__.__module__:
        return array
    elif "jax" in array.__class__.__module__:
        return np.asarray(array)
    else:
        raise TypeError(f"Cannot convert {type(array)} to numpy array")


def trapezoid(y, x=None, dx=1.0, axis=-1):
    """
    A wrapper of `trapz` or `trapezoid` that can handle different
    names in different backends.

    Parameters
    ==========
    y: array-like
        The y values to integrate.
    x: array-like, optional
        The x values to integrate over. If not provided, the spacing
        is assumed to be uniform with spacing `dx`.
    dx: float, optional
        The spacing between x values if `x` is not provided.
    axis: int, optional
        The axis to integrate over.

    Returns
    =======
    float, array-like
        The integral of `y` over `x`.
    """
    if hasattr(xp, "trapezoid"):
        return xp.trapezoid(y, x=x, dx=dx, axis=axis)
    elif hasattr(xp, "trapz"):
        return xp.trapz(y, x=x, dx=dx, axis=axis)
    else:
        raise NotImplementedError(f"trapz/trapezoid not implemented for {xp.__name__}")



"""
Implemented redshift models
"""

import numpy as xp

from ..experimental.cosmo_models import CosmoMixin
from ..utils import trapezoid

__all__ = [
    "_Redshift",
    "PowerLawRedshift",
    "MadauDickinsonRedshift",
    "total_four_volume",
]


class _Redshift(CosmoMixin):
    r"""
    This assumes the model is defined as

    .. math::

        p(z | \Lambda) = \frac{1}{(1 + z)} \frac{dVc}{dz} \psi(z | \Lambda).

    Where :math:`\psi(z | \Lambda)` is implemented in :func:`psi_of_z`.

    Parameters
    ----------
    z_max: float
        The maximum redshift allowed, this is also used to normalize the
        probability.
    cosmo_model: str
        The cosmology model to use. Default is :code:`Planck15`.
        Should be of :code:`wcosmo.available.keys()`.

    Attributes
    ----------
    base_variable_names: list
        :math:`\Lambda` - list of astrophysical rate-evolution parameters
        for the model.
    """

    base_variable_names = None

    @property
    def variable_names(self):
        """
        Variable names for the model

        Returns
        -------
        vars: list
            Variable names including astrophysical rate-evolution parameters
            and cosmological parameters
            (:code:`self.base_variable_names + self.cosmology_names`).
        """
        vars = self.cosmology_names.copy()
        if self.base_variable_names is not None:
            vars += self.base_variable_names
        return vars

    def __init__(self, z_max=2.3, cosmo_model="Planck15"):
        super().__init__(cosmo_model=cosmo_model)
        self.z_max = z_max
        self.zs = xp.linspace(1e-6, z_max, 2500)

    def __call__(self, dataset, **kwargs):
        """
        Wrapper to :func:`probability`.
        """
        return self.probability(dataset=dataset, **kwargs)

    def normalisation(self, parameters):
        r"""
        Compute the normalization of the rate-weighted spacetime volume.

        .. math::

            \mathcal{V} = \int_{0}^{z_{\max}} dz \frac{1}{1+z} \frac{dVc}{dz} \psi(z|\Lambda)

        Here, :math:`z_{\max}` is :code:`self.z_max`.

        Parameters
        ----------
        parameters: dict
            Dictionary of parameters

        Returns
        -------
        norm: float | array-like:
            Total rate-weighted comoving spacetime volume
        """
        normalisation_data = self.differential_spacetime_volume(
            dict(redshift=self.zs), bounds=True, **parameters
        )
        norm = trapezoid(normalisation_data, self.zs)
        return norm

    def probability(self, dataset, **parameters):
        """
        Compute the normalized probability of a merger occurring at the
        specified redshift.

        Parameters
        ----------
        dataset: dict[str, array-like]
            Dictionary of sample points that contains :code:`redshift`
            as a key.
        parameters: dict[str, float]
            Dictionary of parameters :math:`\Lambda`.

        Returns
        -------
        array-like
            The normalized probability per sample
        """
        normalisation = self.normalisation(parameters=parameters)
        differential_volume = self.differential_spacetime_volume(
            dataset=dataset, bounds=True, **parameters
        )
        return differential_volume / normalisation

    def psi_of_z(self, redshift, **parameters):
        r"""
        Method encoding the redshift evolution of the merger rate.
        This should be overwritten in child classes.
        By convention this should be normalized such that

        .. math::

            \psi(0 | \Lambda) = 1.

        Parameters
        ----------
        redshift: array-like
            The redshifts at which to evaluate the model.
        parameters: dict[str, float]
            Dictionary of parameters :math:`\Lambda`.

        Returns
        -------
        array-like
            :math:`\psi(z | \Lambda)` for in input redshifts.
        """
        raise NotImplementedError

    def dvc_dz(self, redshift, **parameters):
        r"""

        .. note::

            The units of this differ from
            :code:`wcosmo.differential_comoving_volume` and
            :code:`astropy.cosmology.FlatwCDM.differential_comoving_volume`
            by a factor of :math:`4 \pi`.

        Returns
        -------
        float
            The differential comoving volume in :math:`{\rm Mpc}^3`
        """
        return (
            4
            * xp.pi
            * self.cosmology(parameters).differential_comoving_volume(redshift)
        )

    def differential_spacetime_volume(self, dataset, bounds=False, **parameters):
        r"""
        Compute the differential spacetime volume.

        .. math::

            d\mathcal{V} = \frac{1}{1+z} \frac{dVc}{dz} \psi(z|\Lambda)

        Parameters
        ----------
        dataset: dict
            Dictionary containing entry "redshift"
        parameters: dict
            Dictionary of parameters

        Returns
        -------
        differential_volume: (float, array-like)
            Differential spacetime volume
        """
        psi_of_z = self.psi_of_z(redshift=dataset["redshift"], **parameters)
        differential_volume = psi_of_z / (1 + dataset["redshift"])
        differential_volume *= self.dvc_dz(redshift=dataset["redshift"], **parameters)
        if bounds:
            differential_volume *= dataset["redshift"] <= self.z_max

        return differential_volume


class PowerLawRedshift(_Redshift):
    base_variable_names = ["lamb"]

    def psi_of_z(self, redshift, **parameters):
        r"""
        Redshift model from Fishbach+ https://arxiv.org/abs/1805.10270
        (`arXiv:1805.10270 <https://arxiv.org/abs/1805.10270>`_.

        .. math::

            \psi(z|\gamma, \kappa, z_p) = (1 + z)^\lambda

        Parameters
        ----------
        lamb: float
            The spectral index :math:`\lambda`.
        """
        return (1 + redshift) ** parameters["lamb"]


class MadauDickinsonRedshift(_Redshift):

    base_variable_names = ["gamma", "kappa", "z_peak"]

    def psi_of_z(self, redshift, **parameters):
        r"""
        Redshift model from Fishbach+
        (`arXiv:1805.10270 <https://arxiv.org/abs/1805.10270>`_ Eq. (33))
        See Callister+ (`arXiv:2003.12152 <https://arxiv.org/abs/2003.12152>`_
        Eq. (2)) for the normalisation.

        .. math::

            \psi(z|\gamma, \kappa, z_p) = \frac{(1 + z)^\gamma}{1 + (\frac{1 + z}{1 + z_p})^\kappa}

        Parameters
        ----------
        gamma: float
            Slope of the distribution at low redshift, :math:`\gamma`.
        kappa: float
            Slope of the distribution at high redshift, :math:`\kappa`.
        z_peak: float
            Redshift at which the distribution peaks, :math:`z_p`.
        """
        gamma = parameters["gamma"]
        kappa = parameters["kappa"]
        z_peak = parameters["z_peak"]
        psi_of_z = (1 + redshift) ** gamma / (
            1 + ((1 + redshift) / (1 + z_peak)) ** kappa
        )
        psi_of_z *= 1 + (1 + z_peak) ** (-kappa)
        return psi_of_z


def total_four_volume(lamb, analysis_time, max_redshift=2.3):
    r"""
    Calculate the rate-weighted four-volume for a given power-law redshift model.

    .. math::

        \mathcal{V} = T \int_{0}^{z_{\max}} dz \frac{dV_c}{dz} (1 + z)^{\lambda - 1}

    Parameters
    ----------
    lamb: float
        The spectral index, :math:`\Lambda`.
    analysis_time: float
        The total analysis time, :math:`T`.
    max_redshift: float, optional
        The maximum redshift to integrate to, :math:`z_{\max}`, default=2.3.

    Returns
    -------
    total_volume: float
        The rate-weighted four-volume

    Notes
    -----
    This assumes a :code:`Planck15` cosmology.
    """
    from wcosmo.astropy import Planck15
    from wcosmo.utils import disable_units

    disable_units()

    redshifts = xp.linspace(0, max_redshift, 2500)
    psi_of_z = (1 + redshifts) ** lamb
    normalization = 4 * xp.pi / 1e9 * analysis_time
    total_volume = (
        trapezoid(
            Planck15.differential_comoving_volume(redshifts)
            / (1 + redshifts)
            * psi_of_z,
            redshifts,
        )
        * normalization
    )
    return total_volume



import numpy as np
import pytest
from bilby.core.prior import PriorDict, Uniform
from wcosmo.astropy import Planck15
from wcosmo.utils import disable_units

import gwpopulation
from gwpopulation.models import redshift
from gwpopulation.utils import trapezoid

N_TEST = 100


def _run_model_normalisation(model, priors, xp=np):
    zs = xp.linspace(1e-3, 2.3, 1000)
    test_data = dict(redshift=zs)
    norms = list()
    for _ in range(N_TEST):
        p_z = model(test_data, **priors.sample())
        norms.append(float(trapezoid(p_z, zs)))
    assert np.max(np.abs(np.array(norms) - 1)) < 1e-3


def test_powerlaw_normalised(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = redshift.PowerLawRedshift()
    priors = PriorDict()
    priors["lamb"] = Uniform(-15, 15)
    _run_model_normalisation(model=model, priors=priors, xp=xp)


def test_madau_dickinson_normalised(backend):
    gwpopulation.set_backend(backend)
    xp = gwpopulation.utils.xp
    model = redshift.MadauDickinsonRedshift()
    priors = PriorDict()
    priors["gamma"] = Uniform(-15, 15)
    priors["kappa"] = Uniform(-15, 15)
    priors["z_peak"] = Uniform(0, 5)
    _run_model_normalisation(model=model, priors=priors, xp=xp)


def test_powerlaw_volume(backend):
    """
    Test that the total volume matches astropy for a
    trivial case
    """
    gwpopulation.set_backend(backend)
    disable_units()
    xp = gwpopulation.utils.xp
    zs = xp.linspace(1e-3, 2.3, 1000)
    zs_numpy = gwpopulation.utils.to_numpy(zs)
    model = redshift.PowerLawRedshift()
    parameters = dict(lamb=1)
    total_volume = trapezoid(
        Planck15.differential_comoving_volume(zs) * 4 * np.pi,
        zs,
    )
    approximation = float(model.normalisation(parameters))
    assert abs(total_volume - approximation) / total_volume < 1e-2


def test_zero_outside_domain():
    model = redshift.PowerLawRedshift(z_max=1)
    assert model(dict(redshift=5), lamb=1) == 0


def test_four_volume():
    disable_units()
    assert (
        Planck15.comoving_volume(2.3) / 1e9
        - redshift.total_four_volume(lamb=1, analysis_time=1, max_redshift=2.3)
        < 1e-3
    )
